"""Helper functions."""

import time
from typing import Any, Iterable

from netports.static import BRIEF_ALL_I, BRIEF_ALL_S, SPLITTER
from netports.types_ import LStr, StrInt, IStrInt, LInt


# =============================== bool ===============================


def is_all(**kwargs) -> bool:
    """Get param `all` from `kwargs`."""
    return bool(kwargs.get("all"))


def is_brief(**kwargs) -> bool:
    """Get param `verbose` from `kwargs` and reverse the result."""
    return not is_verbose(**kwargs)


def is_brief_in_items(items: Any) -> bool:
    """Check is "-1" or -1 in `items`, used for verbose=False.

    :param items: str, int, List[str], List[int].
    :return: True - if "-1" or -1 present in `items`.
    """
    if isinstance(items, (str, int)):
        items = [items]
    if not isinstance(items, Iterable):
        raise TypeError(f"{items=} expected {str}")
    for item in items:
        if isinstance(item, str):
            if item == BRIEF_ALL_S:
                return True
            continue
        if isinstance(item, int):
            if item == BRIEF_ALL_I:
                return True
            continue
        raise TypeError(f"{item=} expected {str}")
    return False


def is_verbose(**kwargs) -> bool:
    """Get param `verbose` from `kwargs`."""
    verbose = kwargs.get("verbose")
    if verbose is None:
        return False
    return bool(verbose)


def is_strict(**kwargs) -> bool:
    """Get param `strict` from `kwargs`."""
    strict = kwargs.get("strict")
    if strict is None:
        return True
    return bool(strict)


# =============================== int ================================


def to_int(number: StrInt) -> int:
    """Convert str to int.

    :param number: int or digit as str/
    :return: int.

    :raises TypeError: If number is not digit.
    """
    if isinstance(number, int):
        return number
    if isinstance(number, str) and number.isdigit():
        return int(number)
    raise TypeError(f"{number=} {int} expected")


def to_lint(numbers: IStrInt) -> LInt:
    """Convert List[str] to List[int].

    :param numbers: List[str].
    :return: List[int].
    :raises TypeError: If number is not digit.
    """
    if not isinstance(numbers, (list, set, tuple)):
        raise TypeError(f"{numbers=} {list} expected")
    return [to_int(s) for s in numbers]


# =============================== list ===============================


def lstr(items: Any) -> LStr:
    """Convert any types of items to List[str]."""
    if not isinstance(items, (bytes, dict, float, int, list, set, str, tuple)):
        raise TypeError(f"{items=} {list} expected")
    if isinstance(items, (bytes, float, int, str)):
        items = [items]
    items_ = []
    for item in items:
        if isinstance(item, bytes):
            items_.append(item.decode("utf-8"))
        elif isinstance(item, float):
            items_.append(str(int(item)))
        else:
            items_.append(str(item))
    return items_


def remove_brief_items(items: Any) -> list:
    """Remove "-1", -1 from items.

    :param items: str, int, List[str], List[int].
    :return: Items without "-1", -1.
    """
    if isinstance(items, (str, int)):
        items = [items]
    return [i for i in items if i not in (BRIEF_ALL_S, BRIEF_ALL_I)]


def split(items: Any) -> LStr:
    """Split items str, List[str] by "," " "."""
    results: LStr = []
    for items_s in lstr(items):
        for item_s in items_s.split():
            results.extend(item_s.split(SPLITTER))
    return results


# ============================= wrapper ==============================


def time_spent(func):
    """Wrapper measure function execution time."""

    def wrap(*args, **kwargs):
        """Wrap."""
        started = time.time()
        return_ = func(*args, **kwargs)
        elapsed = time.time() - started
        print(f"=== {func.__name__}, spent {elapsed:.3f}s ===")
        return return_

    return wrap



"""Ports"""

import re
from typing import Any

from netports import helpers as h
from netports.range import Range
from netports.static import RANGE_SPLITTER, SPLITTER
from netports.types_ import LStr, LInt


# noinspection PyIncorrectDocstring
def parse_range(line: str, **kwargs) -> Range:
    """Parse range from line. Removes white spaces considering splitters.

    Sort numbers and removes duplicates.

    :param line: Range of numbers, can be unsorted and with duplicates,
    :type line: str or List[int] or List[str]

    :param splitter: Separator character between items. Default is ",".
    :type splitter: str

    :param range_splitter: Separator between min and max numbers in range. Default is "-".
    :type range_splitter: str

    :return: Range object of unique sorted numbers.
    :rtype: Range

    :example: Remove duplicates and sort range of numbers.
        parse_range("3-5,1,3-5,1") -> Range("1,3-5")
    """
    splitter = str(kwargs.get("splitter") or SPLITTER)
    range_splitter = str(kwargs.get("range_splitter") or RANGE_SPLITTER)
    regex = rf"(\s+)?{range_splitter}(\s+)?"
    items_: LStr = h.lstr(line)
    items_ = [re.sub(regex, range_splitter, s) for s in items_]
    items_ = [s.replace(range_splitter, "__range__") for s in items_]
    items_ = [s.replace(splitter, "__splitter__") for s in items_]
    items_ = ["".join(s.split()) for s in items_ if s]
    items_ = [s for s in items_ if s]
    items_ = [s.replace("__splitter__", splitter) for s in items_]
    items_ = [s.replace("__range__", range_splitter) for s in items_]
    line_ = splitter.join(items_)
    range_o = Range(items=line_, splitter=splitter, range_splitter=range_splitter)
    return range_o


# noinspection PyIncorrectDocstring
def inumbers(items: Any, **kwargs) -> LInt:
    """Sort integer numbers and removes duplicates.

    :param items: Range of numbers, can be unsorted and with duplicates.
    :type items: str or List[int] or List[str]

    :param splitter: Separator character between items. Default is ",".
    :type splitter: str

    :param range_splitter: Separator between min and max numbers in range. Default is "-".
    :type range_splitter: str

    :return: List of unique sorted numbers.
    :rtype: List[int]

    :example: Converts List[int] to List[int] and removes duplicates.
        inumbers([5, 5, 3, 4, 1]) -> [1, 3, 4, 5]

    :example: Converts List[int] to List[int] and removes duplicates.
        inumbers("1, 7-9, 3 - 5") -> [1, 3, 4, 5, 7, 8, 9]
    """
    range_o: Range = parse_range(items, **kwargs)
    return range_o.numbers()


# noinspection PyIncorrectDocstring
def snumbers(items: Any, **kwargs) -> str:
    """Sort string numbers and removes duplicates.

    :param items: Range of numbers, can be unsorted and with duplicates.
    :type items: str or List[int] or List[str]

    :param splitter: Separator character between items. Default is ",".
    :type splitter: str

    :param range_splitter: Separator between min and max numbers in range. Default is "-".
    :type range_splitter: str

    :return: String of unique sorted numbers.
    :rtype: str

    :example: Converts List[int] to str and removes duplicates.
        snumbers([5, 5, 3, 4, 1]) -> "1,3-5"

    :example: Converts str to str and removes duplicates.
        snumbers("1, 7-9, 3 - 5") -> "1,3-5,7-9"
    """
    range_o: Range = parse_range(items, **kwargs)
    return str(range_o)



"""netports"""

from netports.exceptions import NetportsValueError
from netports.intf import (
    Intf,
    is_port_base,
    sort_names,
)
from netports.intf_gm import (
    IntfGM,
    generate_intfs,
    generate_names,
    names_to_range,
    range_to_intfs,
    range_to_names,
)
from netports.intf_map import (
    long_to_long,
    long_to_short,
    longs,
    short_to_long,
    short_to_short,
    shorts,
)
from netports.ip import IP_NAMES, IP_NUMBERS, iip, sip, ip_pairs
from netports.ipv4 import IPv4
from netports.item import Item
from netports.mac import Mac
from netports.ports import inumbers, parse_range, snumbers
from netports.range import Range
from netports.swversion import SwVersion
from netports.tcp import stcp, itcp, check_port, check_ports
from netports.vlan import ivlan, svlan

__all__ = [
    "IP_NAMES",
    "IP_NUMBERS",
    "IPv4",
    "Intf",
    "IntfGM",
    "Item",
    "Mac",
    "NetportsValueError",
    "Range",
    "SwVersion",
    "check_port",
    "check_ports",
    "generate_intfs",
    "generate_names",
    "iip",
    "inumbers",
    "ip_pairs",
    "is_port_base",
    "itcp",
    "ivlan",
    "long_to_long",
    "long_to_short",
    "longs",
    "names_to_range",
    "parse_range",
    "range_to_intfs",
    "range_to_names",
    "short_to_long",
    "short_to_short",
    "shorts",
    "sip",
    "snumbers",
    "sort_names",
    "stcp",
    "svlan",
]



"""Range, object that represents ports range as str and as List[int]."""

from __future__ import annotations

from functools import total_ordering

from netports import helpers as h
from netports.exceptions import NetportsValueError
from netports.item import Item, LItem
from netports.static import SPLITTER, RANGE_SPLITTER
from netports.types_ import LStr, LInt, OInt, IInt, T2SInt, StrInt, StrIInt, IStrInt


@total_ordering
class Range:
    """Range, object that represents ports range as str and as List[int]."""

    def __init__(self, items: StrIInt = "", **kwargs):
        """Initialize Range.

        :param items: Range of numbers. Numbers can be unsorted and duplicated.
        :type items: str or List[int]

        :param splitter: Separator character between numbers (default ",").
        :type splitter: str

        :param range_splitter: Separator between min and max digits in range (default "-").
        :type range_splitter: str

        :param strict: True - Raise ValueError, if in line is invalid item (default),
            False - Make Range without invalid items.
        :type strict: bool
        """
        self.splitter = kwargs.get("splitter") or SPLITTER
        self.range_splitter = kwargs.get("range_splitter") or RANGE_SPLITTER
        self._strict = self._init_strict(**kwargs)
        self.items: LItem = []
        self._idx = 0

        if isinstance(items, str):
            self.line = items
        elif isinstance(items, int):
            self.items = self._create_items([items])
        elif isinstance(items, (list, set, tuple)):
            self.items = self._create_items(items)
        else:
            raise TypeError(f"{items=} {str} expected")

    # ======================= special methods ========================

    def __repr__(self) -> str:
        """Representation of the object."""
        splitter = self.splitter
        range_splitter = self.range_splitter
        params = [
            f"{self.line!r}",
            f"{splitter=!r}" if splitter != SPLITTER else "",
            f"{range_splitter=!r}" if range_splitter != RANGE_SPLITTER else "",
        ]
        params_ = ", ".join([s for s in params if s])
        class_ = self.__class__.__name__
        return f"{class_}({params_})"

    def __str__(self) -> str:
        """String representation."""
        return self.line

    def __hash__(self) -> int:
        """Hash value of the object."""
        return tuple(self.items).__hash__()

    def __eq__(self, other) -> bool:
        """Check if two objects are equal.

        :param other: Another object to compare.
        :return: True if objects are equal, False otherwise.
        """
        if self.__class__ == other.__class__:
            if self.__hash__() == other.__hash__():
                return True
        return False

    def __lt__(self, other: Range) -> bool:
        """Compare two objects.

        :param other: Another object to compare with.
        """
        if self.__class__ == other.__class__:
            self_len = len(self.items)
            other_len = len(other.items)
            if not (self_len and other_len):
                return self_len < other_len
            for idx, self_item in enumerate(self.items):
                try:
                    other_item = other.items[idx]
                except IndexError:
                    return False
                if self_item.min != other_item.min:
                    return self_item.min < other_item.min
                if self_item.max != other_item.max:
                    return self_item.max < other_item.max
            return self.line < other.line
        return False

    def __add__(self, other: Range) -> Range:
        """+ Add."""
        self_numbers, other_numbers = self._numbers_sets(other)
        self_numbers.update(other_numbers)
        return Range(self_numbers)

    def __sub__(self, other: Range) -> Range:
        """- Subtract."""
        self_numbers, other_numbers = self._numbers_sets(other)
        self_numbers.difference_update(other_numbers)
        return Range(self_numbers)

    def __contains__(self, number: int) -> bool:
        """Return key in self."""
        if not isinstance(number, int):
            raise TypeError
        return number in self.numbers()

    def __delitem__(self, idx: int) -> None:
        """Delete self.numbers[idx]."""
        numbers = self.numbers()
        numbers.__delitem__(idx)
        self.items = self._create_items(numbers)

    def __getitem__(self, idx: int):
        """Return number by index."""
        return self.numbers()[idx]

    def __iter__(self):
        """Iterator."""
        return self

    def __len__(self) -> int:
        """Return length of numbers."""
        return len(self.numbers())

    def __next__(self) -> int:
        """Return next number."""
        try:
            number = self.numbers()[self._idx]
        except IndexError as ex:
            raise StopIteration() from ex
        self._idx += 1
        return number

    # ======================= list/set methods =======================

    def add(self, other: Range) -> None:
        """Add other Range object to self."""
        self_numbers, other_numbers = self._numbers_sets(other)
        self_numbers.update(other_numbers)
        self.items = self._create_items(self_numbers)

    def append(self, number: StrInt) -> None:
        """Append number to self."""
        number_ = h.to_int(number)
        other = Range(number_)
        self.add(other)

    def clear(self) -> None:
        """Remove all numbers from self."""
        self.items = []

    def copy(self):
        """Return a copy of self Range object."""
        return Range(
            items=self.line,
            splitter=self.splitter,
            range_splitter=self.range_splitter,
            strict=self._strict,
        )

    def difference(self, other: Range) -> Range:
        """Return the Range object of the difference between self and other Range."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.difference(other_numbers)
        return Range(numbers)

    def difference_update(self, other: Range) -> None:
        """Remove other Range from self."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.difference(other_numbers)
        self.items = self._create_items(numbers)

    def discard(self, number: StrInt) -> None:
        """Remove the specified number from self Range."""
        number_ = h.to_int(number)
        numbers = set(self.numbers()).difference({number_})
        self.items = self._create_items(numbers)

    def extend(self, numbers: IInt) -> None:
        """Add List[int] numbers to self."""
        if not isinstance(numbers, (list, set, tuple)):
            raise TypeError(f"{numbers=} {list} expected")
        other = Range(numbers)
        self.add(other)

    def index(self, number: StrInt) -> int:
        """Index of number.

        :return: Returns index of number.
        :raises ValueError: if the number is not present in range.
        """
        number_ = h.to_int(number)
        return self.numbers().index(number_)

    def intersection(self, other: Range) -> Range:
        """Return Range which is the intersection of self and other Range."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.intersection(other_numbers)
        return Range(numbers)

    def intersection_update(self, other: Range) -> None:
        """Remove numbers of other Range in self, that are not present in other."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.intersection(other_numbers)
        self.items = self._create_items(numbers)

    def isdisjoint(self, other: Range) -> bool:
        """Return whether self numbers and other Range numbers have intersection or not."""
        self_numbers, other_numbers = self._numbers_sets(other)
        return self_numbers.isdisjoint(other_numbers)

    def issubset(self, other: Range) -> bool:
        """Return whether other Range numbers contains self numbers or not."""
        self_numbers, other_numbers = self._numbers_sets(other)
        return self_numbers.issubset(other_numbers)

    def issuperset(self, other: Range) -> bool:
        """Return whether self Range numbers contains other Range numbers set or not."""
        self_numbers, other_numbers = self._numbers_sets(other)
        return self_numbers.issuperset(other_numbers)

    def pop(self) -> int:
        """Remove and returns last number in Range.

        :raises IndexError: If list is empty or index is out of range.
        """
        numbers = self.numbers()
        number = numbers.pop()
        self.items = self._create_items(numbers)
        return number

    def remove(self, number: StrInt) -> None:
        """Remove the specified number from self Range.

        :raises ValueError: If the numbers is not present.
        """
        number_ = h.to_int(number)
        numbers = self.numbers()
        numbers.remove(number_)
        self.items = self._create_items(numbers)

    def symmetric_difference(self, other: Range) -> Range:
        """Return Range object with the symmetric differences of self and other Range."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.symmetric_difference(other_numbers)
        return Range(numbers)

    def symmetric_difference_update(self, other: Range) -> None:
        """Insert the symmetric differences from self Range and other Range."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.symmetric_difference(other_numbers)
        self.items = self._create_items(numbers)

    def union(self, other: Range) -> Range:
        """Return Range of the union of self and other numbers."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.union(other_numbers)
        return Range(numbers)

    def update(self, other: Range) -> None:
        """Return Range of the union of self Range and other Range."""
        self_numbers, other_numbers = self._numbers_sets(other)
        numbers = self_numbers.union(other_numbers)
        self.items = self._create_items(numbers)

    # ============================= init =============================

    @staticmethod
    def _init_strict(**kwargs) -> bool:
        """Init strict."""
        strict = kwargs.get("strict")
        if strict is None:
            return True
        if isinstance(strict, bool):
            return strict
        raise TypeError(f"{strict=} {bool} expected")

    # =========================== property ===========================

    @property
    def line(self) -> str:
        """Range in str format."""
        return self._items_to_line(self.items)

    @line.setter
    def line(self, line: str) -> None:
        if not isinstance(line, str):
            raise TypeError(f"{line=} {str} expected")
        line_: str = self._valid_line(line)
        line_ = line_.replace(self.range_splitter, RANGE_SPLITTER)
        line_ = line_.replace(self.splitter, SPLITTER)
        lines: LStr = line_.split(SPLITTER)
        items = self._create_items(lines)
        self.items = items

    # =========================== methods ============================

    def numbers(self) -> LInt:
        """Return list of numbers."""
        return [i for o in self.items for i in o.range]

    # =========================== helpers ============================

    def _items_to_line(self, items: LItem) -> str:
        """Convert  items *List[Item]* to line str.

        :param items: [Item("1"), Item("3-5")].
        :return: "1,3-5".
        """
        lines = [o.line.replace("-", self.range_splitter) for o in items]
        line = self.splitter.join(lines)
        return line

    @staticmethod
    def _items_wo_duplicates(items: LItem) -> LItem:
        """Remove duplicates digits in items.

        :param items: [Item(1), Item(4-5), Item(3-4), Item(1)].
        :return: [Item(1), Item(3-5)].
        """
        items_: LStr = []  # result
        numbers: LInt = sorted({i for o in items for i in o.range})
        item_1st: OInt = None
        for idx, item in enumerate(numbers, start=1):
            # not last iteration
            if idx < len(numbers):
                item_next = numbers[idx]
                if item_next - item <= 1:  # range
                    if item_1st is None:  # start new range
                        item_1st = item
                else:  # int or end of range
                    if item_1st is None:
                        range_ = str(item)
                    else:
                        range_ = f"{item_1st}-{item}"
                    items_.append(range_)
                    item_1st = None
            # last iteration
            else:
                item_ = str(item) if item_1st is None else f"{item_1st}-{item}"
                items_.append(item_)
        return [Item(s) for s in items_]

    def _numbers_sets(self, other: Range) -> T2SInt:
        """Convert self and other List[int] numbers to Set[int]

        :param other: Other Range object.
        :return: Sets of numbers.
        """
        if not isinstance(other, Range):
            raise TypeError(f"{other=} {Range} expected")
        self_numbers = self.numbers()
        other_numbers = other.numbers()
        return set(self_numbers), set(other_numbers)

    def _create_items(self, items: IStrInt) -> LItem:
        """Convert items List[str] to items List[Items], removes duplicates.

        :param items: List of str items.
        :return: List of *Item* objects.
        :raises ValueError: If self._strict==True and item is invalid.
        """
        items_: LItem = []
        items_wo_duplicates = sorted({str(i) for i in items})
        for item in items_wo_duplicates:
            if item == "":
                continue
            try:
                items_.append(Item(str(item)))
            except ValueError as ex:
                if self._strict:
                    raise type(ex)(*ex.args)
                continue
        items_ = self._items_wo_duplicates(items_)
        return items_

    def _valid_line(self, line: str) -> str:
        """Check valid chars in line. Splits line to items by splitter."""
        lines: LStr = []
        splitter = self.splitter
        range_splitter = self.range_splitter
        lines_ = line.split(range_splitter)
        for line_ in lines_:
            items_: LStr = line_.split(splitter)
            items_ = [s for s in items_ if s]
            for item in items_:
                if not item.isdigit():
                    if not self._strict:
                        continue
                    raise NetportsValueError(f"{item=} in {line=}")
            line_ = splitter.join(items_)
            lines.append(line_)
        return range_splitter.join(lines)



"""MAC address representation in different formats."""

from __future__ import annotations

import string
from functools import total_ordering
from typing import List, Dict

from pydantic import BaseModel, Field

from netports.exceptions import NetportsValueError
from netports.types_ import LStr, T2Str


@total_ordering
class Mac(BaseModel):
    """MAC address representation in different formats."""

    addr: str = Field(description="MAC address")

    def __init__(self, *args, **kwargs):
        """Initialize Mac.

        :param addr: MAC address.
        :raises NetportsValueError: If the MAC address is invalid or empty.
        """
        addr, hex_ = _validate_addr(*args, **kwargs)
        kwargs_ = {k: v for k, v in kwargs.items() if k != "addr"}
        super().__init__(addr=addr, **kwargs_)
        self._hex = hex_

    def __repr__(self) -> str:
        """Representation of the object."""
        class_ = self.__class__.__name__
        return f"{class_}({self.addr!r})"

    def __str__(self) -> str:
        """String representation."""
        return str(self.addr)

    def __hash__(self) -> int:
        """Hash value of the object."""
        return hash(self.integer)

    def __eq__(self, other) -> bool:
        """Check if two objects are equal.

        :param other: Another object to compare.
        :return: True if objects are equal, False otherwise.
        """
        if self.__class__ == other.__class__:
            if self.__hash__() == other.__hash__():
                return True
        return False

    def __lt__(self, other) -> bool:
        """Compare two objects.

        :param other: Another object to compare with.
        """
        if self.__class__ == other.__class__:
            return self.integer < other.integer
        return False

    def __copy__(self) -> Mac:
        """Create a duplicate of the object.

        :return: A copy of the current object.
        """
        return type(self)(line=self.addr)

    # ============================= property =============================

    @property
    def hex(self) -> str:
        """MAC address as 12-character hexadecimal string."""
        return self._hex

    @property
    def integer(self) -> int:
        """MAC address as integer."""
        return int(self.hex, 16)

    @property
    def cisco(self) -> str:
        """Return MAC in Cisco dot format: 0000.0000.0000."""
        addr = self.format(size=4, splitter=".")
        return addr.lower()

    @property
    def hp_comware(self) -> str:
        """Return MAC in hp_comware dash format: 0000-0000-0000."""
        addr = self.format(size=4, splitter="-")
        return addr.lower()

    @property
    def hp_procurve(self) -> str:
        """Return MAC in hp_procurve dash format: 000000-000000."""
        addr = self.format(size=6, splitter="-")
        return addr.lower()

    @property
    def linux(self) -> str:
        """Return MAC in Linux colon format: 00:00:00:00:00:00."""
        addr = self.format(size=2, splitter=":")
        return addr.lower()

    @property
    def windows(self) -> str:
        """Return MAC in Windows dash format: 00-00-00-00-00-00."""
        addr = self.format(size=2, splitter="-")
        return addr.lower()

    # ============================= methods ==============================

    def format(self, size: int, splitter: str) -> str:
        """Format MAC address, split into chunks of given size.

        :param size: Integer value to determine the grouping size.
        :param splitter: String to use as a separator.
        :return: Formatted string.
        :raises NetportsValueError: If invalid size.
        """
        if size not in [2, 4, 6]:
            raise NetportsValueError(f"Invalid {size=}")
        chunks = [self.hex[i : i + size] for i in range(0, 12, size)]
        return splitter.join(chunks)


LMac = List[Mac]
DMac = Dict[str, Mac]


# ============================= helpers ==============================


def _validate_addr(*args, **kwargs) -> T2Str:
    """Extract MAC address line from arguments or keyword arguments.

    :param args: Tuple of arguments.
    :param kwargs: Dictionary of keyword arguments.
    :return: MAC address as a string.
    """
    # addr
    addr = ""
    if args:
        addr = str(args[0])
    if not addr:
        addr = str(kwargs.get("addr") or "")
    if not addr:
        raise NetportsValueError("MAC address is empty.")

    # splitter
    expected = string.hexdigits + ":.-"
    if set(addr).difference(set(expected)):
        raise NetportsValueError("Invalid MAC address splitter")

    # hex
    hexdigits: LStr = [s for s in addr if s in string.hexdigits]
    line = "".join(hexdigits)
    if len(line) != 12:
        raise NetportsValueError("Invalid MAC address format, 12 hexdigits expected.")
    hex_ = "".join(hexdigits)

    return addr, hex_



"""Network interface representation with up to 6 numerical indices."""

import re
from functools import total_ordering
from typing import List, Optional, Set, Tuple, Union

from vhelpers import vlist

from netports import intf_map
from netports.exceptions import NetportsValueError
from netports.types_ import T5Str, LStr, SStr, DStr, T7Str, OLT2Str, OLStr

DEVICE_TYPES = [
    "cisco_ios",
    "cisco_nxos",
    "cisco_xr",
    "hp_comware",
    "hp_procurve",
]
SPLITTER = ",./:"


@total_ordering
class Intf:
    """Network interface representation with up to 6 numerical indices."""

    def __init__(self, line: str = "", **kwargs):
        """Initialize Intf.

        :param line: Interface name that can contain up to 4 indexes.
        :type line: str
        :param device_type: Netmiko device type (default "").
        :type device_type: str
        :param splitter: Separator of characters between indexes (default ",./:").
        :type splitter: str
        """
        self._device_type = _init_device_type(**kwargs)
        self._splitter = _init_splitter(**kwargs)
        self._ids: T7Str = self._init_ids(line)
        self._delimiters: T5Str = self._init_delimiters(line)
        self.line = self._init_line()

    def __repr__(self) -> str:
        """Representation of the object."""
        class_ = self.__class__.__name__
        return f"{class_}({self.line!r})"

    def __str__(self) -> str:
        """String representation."""
        return str(self.line)

    def __hash__(self) -> int:
        """Hash value of the object."""
        ids = (self.id0, self.id1, self.id2, self.id3, self.id4, self.id5, self.id6)
        return hash(ids)

    def __eq__(self, other) -> bool:
        """Check if two objects are equal.

        :param other: Another object to compare.
        :return: True if objects are equal, False otherwise.
        """
        if self.__class__ != other.__class__:
            return False
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """Compare two objects.

        :param other: Another object to compare with.
        """
        if self.__class__ == other.__class__:
            if self.id0 != other.id0:
                return self.id0 < other.id0
            if self.id1 != other.id1:
                return self.id1 < other.id1
            if self.id2 != other.id2:
                return self.id2 < other.id2
            if self.id3 != other.id3:
                return self.id3 < other.id3
            if self.id4 != other.id4:
                return self.id4 < other.id4
            if self.id5 != other.id5:
                return self.id5 < other.id5
            return self.id6 < other.id6
        return False

    # =========================== property ===========================

    @property
    def delimiters(self) -> T5Str:
        """Interface all delimiters."""
        return self._delimiters

    @property
    def id0(self) -> str:
        """Interface name. Line without IDs.

        :example:
            Intf("interface Ethernet1/2/3.4").id0 -> "interface Ethernet"
        """
        return self._ids[0]

    @property
    def id1(self) -> int:
        """Interface 1st ID.

        :example:
            Intf("interface Ethernet1/2/3.4").id1 -> 1
        """
        return int(self._ids[1]) if self._ids[1] else 0

    @property
    def id2(self) -> int:
        """Interface 2nd ID.

        :example:
            Intf("interface Ethernet1/2/3.4").id2 -> 2
        """
        return int(self._ids[2]) if self._ids[2] else 0

    @property
    def id3(self) -> int:
        """Interface 3rd ID.

        :example:
            Intf("interface Ethernet1/2/3.4") -> 3
        """
        return int(self._ids[3]) if self._ids[3] else 0

    @property
    def id4(self) -> int:
        """Interface 4th ID.

        :example:
            Intf("interface Ethernet1/2/3.4").id4 -> 4
        """
        return int(self._ids[4]) if self._ids[4] else 0

    @property
    def id5(self) -> int:
        """Interface 5th ID.

        :example:
            Intf("interface Ethernet1/2/3/4.5").id5 -> 5
        """
        return int(self._ids[5]) if self._ids[5] else 0

    @property
    def id6(self) -> int:
        """Interface 6th ID.

        :example:
            Intf("interface Ethernet1/2/3/4/5.6").id6 -> 6
        """
        return int(self._ids[6]) if self._ids[6] else 0

    @property
    def ids(self) -> T7Str:
        """Interface all IDs.

        :example:
            Intf("interface Ethernet1/2/3.4").ids ->
                "interface Ethernet", "1", "2", "3", "4", "", "")
        """
        return self._ids

    @property
    def name(self) -> str:
        """Interface name with IDs.

        :example:
            Intf("interface Ethernet1/2/3.4").name -> "Ethernet1/2/3.4"
        """
        return re.sub(r"^interface(\s+)?", "", self.line)

    @property
    def device_type(self) -> str:
        """Netmiko device type."""
        return self._device_type

    @property
    def splitter(self) -> str:
        """Separator of characters between indexes."""
        return self._splitter

    # =========================== methods ============================

    def all_names(self) -> LStr:
        """All variants of names: long, short, upper-case, lover-case.

        :example:
            Intf("Eth1/2").names() -> [
                "interface Ethernet1/2",
                "interface ethernet1/2",
                "interface Eth1/2",
                "interface eth1/2",
                "Ethernet1/2",
                "ethernet1/2",
                "Eth1/2",
                "eth1/2",
            ]
        """
        results: SStr = set()

        names_: LStr = [self.line, self.name, self.name_full(), self.name_long(), self.name_short()]
        names_ = vlist.no_dupl(names_)
        for name_ in names_:
            results.add(name_)
            if not name_.startswith("interface "):
                name_ = f"interface {name_}"
                results.add(name_)

        intf_map_upper: DStr = intf_map.short_to_long(self._device_type)
        intf_map_lower: DStr = intf_map.short_to_long(self._device_type, key_lower=True)
        names: LStr = [self.name, self.name_short()]
        names = vlist.no_dupl(names)
        for name in names:
            intf_o = Intf(line=name, device_type=self.device_type)
            for id0_short, intf_map_d in [
                (intf_o.id0, intf_map_upper),
                (intf_o.id0.lower(), intf_map_lower),
            ]:
                if id0_long := intf_map_d.get(id0_short) or "":
                    name_long = intf_o.line.replace(id0_short, id0_long, 1)
                    results.add(name_long)
                    if not name_long.startswith("interface "):
                        results.add(f"interface {name_long}")

        results.update([s.lower() for s in results])

        results_: LStr = sorted(results)
        results_.sort(key=len, reverse=True)

        if self.device_type == "hp_procurve":
            if digits := [s for s in results if s.isdigit()]:
                names = [f"interface 1/{s}" for s in digits]
                results_.extend(names)
        return results_

    def last_idx(self) -> int:
        """Index of last ID in interface line.

        :example:
            Intf("interface Ethernet1/2/3.4").last_idx() -> 4
        """
        ids = self._ids[1:]
        return len([s for s in ids if s])

    def name_base(self) -> str:
        """Interface base name without IDs.

        :example:
            Intf("interface Ethernet1/2").name_base() -> "Ethernet"
        """
        return self.id0.replace("interface", "", 1).strip()

    def name_full(self) -> str:
        """Interface long name with IDs and with interface keyword.

        :example:
            Intf("Eth1/2").name_full() -> "interface Ethernet1/2"
        """
        name = self.name_long()
        if not name.startswith("interface"):
            name = " ".join([s for s in ("interface", name) if s])
        return name

    def name_long(self) -> str:
        """Interface long name with IDs and without interface keyword.

        :example:
            Intf("Eth1/2").name_long() -> "Ethernet1/2"
        """
        id0 = self.id0.lower()
        if id0.startswith("interface"):
            id0 = re.sub(r"^interface(\s+)?", "", id0, re.I)

        intf_map_short: DStr = intf_map.short_to_long(self._device_type, key_lower=True)
        for short_lower, long_upper in intf_map_short.items():
            if id0 == short_lower:
                id0 = long_upper
                break
        else:
            intf_map_long: DStr = intf_map.long_to_long(self._device_type, key_lower=True)
            for long_lower, long_upper in intf_map_long.items():
                if id0 == long_lower:
                    id0 = long_upper
                    break

        id1 = self.part_after(idx=0)
        name = f"{id0}{id1}"
        return name.strip()

    def name_short(self, replace: OLT2Str = None) -> str:
        """Interface short name with IDs.

        :param replace: Replace the default short name with the first one
            that matches in the list of the `replace` argument.
        :return: Interface short name.

        :example:
            Intf("interface FastEthernet1/2").name_short() -> "Fa1/2"
            Intf("interface FastEthernet1/2").name_short(replace=[("Fa", "Eth")]) -> "Eth1/2"
        """
        id0 = self.id0.lower()
        if id0.startswith("interface"):
            id0 = re.sub(r"^interface(\s+)?", "", id0, re.I)

        intf_map_l2s: DStr = intf_map.long_to_short(self._device_type, key_lower=True)
        for long_lower, short_upper in intf_map_l2s.items():
            if id0 == long_lower:
                id0 = short_upper
                break
        else:
            intf_map_s2s: DStr = intf_map.short_to_short(self._device_type, key_lower=True)
            for short_lower, short_upper in intf_map_s2s.items():
                if id0 == short_lower:
                    id0 = short_upper
                    break

        if replace:
            for before, after in replace:
                if id0 == before:
                    id0 = after
                    break

        id1 = self.part_after(idx=0)
        name = f"{id0}{id1}"
        return name.strip()

    def part_after(self, idx: int, splitter=True) -> str:
        """Interface part after interested ID.

        :param idx: Interface index.
        :param splitter: True - Include splitter from edge, False - Skip splitter from edge.
        :return: Part of the interface name after specified interface index.

        :example:
            intf = Intf("Ethernet1/2/3.4")
            intf.part_after(0) -> "1/2/3.4"
            intf.part_after(1) -> "2/3.4"
            intf.part_after(2) -> "/3.4"
            intf.part_after(3) -> ".4"
            intf.part_after(2, splitter=False) -> "3.4"
        """
        if idx >= 6:
            return ""

        parts = self._ids[6]
        if idx == 5:
            if splitter:
                return self._delimiters[4] + parts
            return parts

        parts = self._ids[5] + self._delimiters[4] + parts
        if idx == 4:
            if splitter:
                return self._delimiters[3] + parts
            return parts

        parts = self._ids[4] + self._delimiters[3] + parts
        if idx == 3:
            if splitter:
                return self._delimiters[2] + parts
            return parts

        parts = self._ids[3] + self._delimiters[2] + parts
        if idx == 2:
            if splitter:
                return self._delimiters[1] + parts
            return parts

        parts = self._ids[2] + self._delimiters[1] + parts
        if idx == 1:
            if splitter:
                return self._delimiters[0] + parts
            return parts

        parts = self._ids[1] + self._delimiters[0] + parts
        if idx == 0:
            return parts

        parts = self._ids[0] + parts
        return parts

    def part_before(self, idx: int, splitter=True) -> str:
        """Interface part before interested ID.

        :param idx: Interface index.
        :param splitter: True - Include splitter from edge, False - Skip splitter from edge.
        :return: Part of the interface name before specified interface index.

        :example:
            intf = Intf("Ethernet1/2/3.4")
            intf.part_before(2) -> "Ethernet1/"
            intf.part_before(3) -> "Ethernet1/2/"
            intf.part_before(3, splitter=False) -> "Ethernet1/2"
        """
        if idx <= 0:
            return ""

        parts = self._ids[0]
        if idx == 1:
            return parts

        parts += self._ids[1]
        if idx == 2:
            if splitter:
                return parts + self._delimiters[0]
            return parts

        parts += self._delimiters[0] + self._ids[2]
        if idx == 3:
            if splitter:
                return parts + self._delimiters[1]
            return parts

        parts += self._delimiters[1] + self._ids[3]
        if idx == 4:
            if splitter:
                return parts + self._delimiters[2]
            return parts

        parts += self._delimiters[2] + self._ids[4]
        if idx == 5:
            if splitter:
                return parts + self._delimiters[3]
            return parts

        parts += self._delimiters[3] + self._ids[5]
        if idx == 6:
            if splitter:
                return parts + self._delimiters[4]
            return parts

        return self.line

    # =========================== helpers ============================

    def _init_line(self) -> str:
        """Parse Interface line.

        :return: Interface line with IDs and delimiters.
        """
        items: LStr = [
            self._ids[0],
            self._ids[1],
            self._delimiters[0],
            self._ids[2],
            self._delimiters[1],
            self._ids[3],
            self._delimiters[2],
            self._ids[4],
            self._delimiters[3],
            self._ids[5],
            self._delimiters[4],
            self._ids[6],
        ]
        items = [s for s in items if s]
        return "".join(items)

    def _init_ids(self, intf: str) -> T7Str:
        """Split interface line to name and IDs.

        :param intf: Interface line to split.
        :example:
            self.line: "interface Ethernet1/2/3.4"
            return: ("interface Ethernet", "1", "2", "3", "4")
        """
        for splitter in self._splitter:
            intf = intf.replace(splitter, ",")

        name = r"([a-zA-Z\-\s]+)*"
        id1 = r"(\d+)*"
        id2 = r"(?:,)*(\d+)*"
        pattern = f"{name}{id1}{id2}{id2}{id2}{id2}{id2}"
        ids: T7Str = (re.findall(pattern, intf) or [("", "", "", "", "", "", "")])[0]
        return ids

    def _init_delimiters(self, intf: str) -> T5Str:  # pylint: disable=too-many-locals
        """Split interface line to splitters of IDs.

        :param intf: Interface line to split.
        :return: Tuple of up to 5 delimiters between IDs.
        :example:
            self.line: "interface Ethernet1/2/3.4"
            return: ("/", "/", ".", "", "")
        """
        part1 = self._ids[0] + self._ids[1]
        len1 = len(part1)
        delim1 = intf[len1 : len1 + 1]

        part2 = part1 + delim1 + self._ids[2]
        len2 = len(part2)
        delim2 = intf[len2 : len2 + 1]

        part3 = part2 + delim2 + self._ids[3]
        len3 = len(part3)
        delim3 = intf[len3 : len3 + 1]

        part4 = part3 + delim3 + self._ids[4]
        len4 = len(part4)
        delim4 = intf[len4 : len4 + 1]

        part5 = part4 + delim4 + self._ids[5]
        len5 = len(part5)
        delim5 = intf[len5 : len5 + 1]
        return delim1, delim2, delim3, delim4, delim5


LIntf = List[Intf]
SIntf = Set[Intf]
TIntf = Tuple[Intf]
ULIntf = Optional[Union[str, LStr, Intf, LIntf]]


# ============================ functions =============================


def is_port_base(port: str, required: OLStr = None, ignore: OLStr = None) -> bool:
    """Check if the port has one of the required base, skipping base that are in the ignore list.

    :param port: Port name that need to check.
    :type port: str

    :param required: Required base names (without ID),
        one of which should match with port base name.
    :type required: List[str]

    :param ignore: Base names to ignore.
    :type ignore: List[str]

    :return: True - if port base name matches with required, False - otherwise.
    :rtype: bool
    """
    base = Intf(port).id0
    if ignore:
        if base in ignore:
            return True
    if required:
        if base in required:
            return True
    return False


def sort_names(names: LStr, reverse: bool = False) -> LStr:
    """Sort interface names.

    :param names: Interface names that need to be sorted.
    :type names: List[str]

    :param reverse: True - descending, False - ascending, default is False.
    :type reverse: bool

    :return: Sorted interface names.
    :rtype: List[str]
    """
    intfs: LIntf = sorted([Intf(s) for s in names], reverse=reverse)
    return [o.line for o in intfs]


# ============================= helpers ==============================


def _init_device_type(**kwargs) -> str:
    """Validate Netmiko device type.

    :return: Device type.
    :raises NetportsValueError: if the device type is unsupported.
    """
    device_type = str(kwargs.get("device_type") or "")
    expected = ["", *DEVICE_TYPES]
    if device_type not in expected:
        raise NetportsValueError(f"{device_type=} {expected=}")
    return device_type


def _init_splitter(**kwargs) -> str:
    """Validate splitter between interface IDs.

    :return: Splitters pattern.
    """
    if splitter := str(kwargs.get("splitter") or ""):
        return splitter
    return SPLITTER



"""Interface name mapping"""

from netports.types_ import DStr, LStr, SStr, OBool

MAP_OTHER = {  # parent of all
    "At": "ATM",
    "Bu": "BundleEthernet",
    "FD": "Fddi",
    "Hu": "HundredGigabitEthernet",  # overlapped ios
    "Ma": "Management",
    "Mu": "Multilink",
    "Se": "Serial",
    "V": "Vlan",
    "Vi": "Virtual-Access",
    "Vt": "Virtual-Template",
}
MAP_CISCO_IOS = {  # parent of nxos, xr
    "Ap": "AppGigabitEthernet",  # Application Hosting
    "Em": "Embedded-Service-Engine",
    "Eth": "Ethernet",
    "Fa": "FastEthernet",
    "Gi": "GigabitEthernet",
    "Hu": "HundredGigE",  # overlapped: other
    "Lo": "Loopback",  # overlapped: nxos
    "Po": "Port-channel",  # overlapped: nxos
    "Te": "TenGigabitEthernet",  # overlapped: xr
    "Tu": "Tunnel",  # overlapped: xr
    "Twe": "TwentyFiveGigE",
    "Vlan": "Vlan",  # overlapped: h3c
}
MAP_CISCO_NXOS = {
    "Po": "port-channel",  # overlapped: ios
    "Lo": "loopback",  # overlapped: ios, xr
    "mgmt": "mgmt",
}
MAP_CISCO_XR = {
    "Te": "TenGigE",  # overlapped: ios
    "BE": "Bundle-Ether",
    "Tu": "tunnel-ip",  # low-priority, overlapped: ios, nxos
    "ti": "tunnel-ip",
    "Mg": "MgmtEth",
}
MAP_HP_COMWARE = {  # h3c
    "GE": "GigabitEthernet",
    "Te": "Ten-GigabitEthernet",  # low-priority, overlapped: ios
    "XGE": "Ten-GigabitEthernet",
    "BAGG": "Bridge-Aggregation",
    "Vlan": "Vlan-interface",  # overlapped: ios
}
MAP_HP_PROCURVE = {  # hpc
    "Trk": "Trk",
}
ALL_SHORT = sorted(
    {
        *MAP_OTHER,
        *MAP_CISCO_IOS,
        *MAP_CISCO_NXOS,
        *MAP_CISCO_XR,
        *MAP_HP_COMWARE,
        *MAP_HP_PROCURVE,
    }
)


def long_to_short(
    device_type: str = "", key_lower: bool = False, value_lower: bool = False
) -> DStr:
    """Return Interfaces map long-to-short, device_type specific.

    :param device_type: Netmiko device type.
    :param key_lower: True - keys lower-case, False - keys upper-case.
    :param value_lower: True - values lower-case, False - values upper-case.
    :return: Interfaces map.

    :example:
        long_to_short() -> {"Vlan": "V", ...}
        long_to_short(device_type="cisco_ios") -> {"Vlan": "Vlan", ...}
        long_to_short(device_type="cisco_ios", key_lower=True) -> {"vlan": "Vlan", ...}
    """
    data: DStr = short_to_long(device_type=device_type)
    data = {v: k for k, v in data.items()}
    data_ = _overlapped(data)
    data.update(data_)

    if device_type == "cisco_xr":
        data["Tunnel"] = "ti"
    elif device_type == "hp_comware":
        data["TenGigabitEthernet"] = "XGE"

    data = _lower(data, key_lower, value_lower)
    return data


def long_to_long(device_type: str = "", key_lower: bool = False, value_lower: bool = False) -> DStr:
    """Return Interfaces map long-to-long, device_type specific.

    :param device_type: Netmiko device type.
    :param key_lower: True - keys lower-case, False - keys upper-case.
    :param value_lower: True - values lower-case, False - values upper-case.
    :return: Interfaces map.
    """
    data: DStr = short_to_long(device_type=device_type)
    data = {v: k for k, v in data.items()}
    data = {k: k for k in data}
    data = _lower(data, key_lower, value_lower)
    return data


def longs(device_type: str = "", value_lower: OBool = None) -> LStr:
    """Return long names of all interfaces.

    :param device_type: Netmiko device type.
    :param value_lower: True - values lower-case, False - values upper-case.
        Default is None, lower-case and upper-case.
    :return: Long names of all interfaces.

    :example:
        longs() -> ["ATM", "Bridge-Aggregation", "Bundle-Ether", ...]
        longs(device_type="cisco_ios") ->["ATM", "Ethernet", "FastEthernet", ...]
        longs(device_type="cisco_ios", value_lower=True) -> ["atm", "ethernet", ...]
    """
    if value_lower is None:
        upper: SStr = set(long_to_long(device_type=device_type, key_lower=False))
        lower: SStr = set(long_to_long(device_type=device_type, key_lower=True))
        return sorted(upper.union(lower))
    if value_lower:
        return sorted(long_to_long(device_type=device_type, key_lower=True))
    return sorted(long_to_long(device_type=device_type, key_lower=False))


def short_to_long(
    device_type: str = "", key_lower: bool = False, value_lower: bool = False
) -> DStr:
    """Return Interfaces map short-to-long, device_type specific.

    :param device_type: Netmiko device type.
    :param key_lower: True - keys lower-case, False - keys upper-case.
    :param value_lower: True - values lower-case, False - values upper-case.
    :return: Interfaces map.

    :example:
        short_to_long() -> {"Fa": "FastEthernet", ...}
        short_to_long(key_lower=True) -> {"fa": "FastEthernet", ...}
        short_to_long(value_lower=True) -> {"Fa": "fastethernet", ...}
    """
    if device_type == "cisco_xr":
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_CISCO_XR)
    elif device_type == "cisco_ios":
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
    elif device_type == "cisco_nxos":
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_CISCO_NXOS)
    elif device_type == "hp_comware":  # h3c
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_HP_COMWARE)
    elif device_type == "hp_procurve":  # h3c
        data = MAP_OTHER.copy()
        data.update(MAP_CISCO_IOS)
        data.update(MAP_HP_PROCURVE)
    else:
        data = MAP_HP_COMWARE.copy()
        data.update(MAP_HP_PROCURVE)
        data.update(MAP_CISCO_XR)
        data.update(MAP_CISCO_NXOS)
        data.update(MAP_CISCO_IOS)
        data.update(MAP_OTHER)
    data = _lower(data, key_lower, value_lower)
    return data


def short_to_short(
    device_type: str = "", key_lower: bool = False, value_lower: bool = False
) -> DStr:
    """Return Interfaces map short-to-short, device_type specific.

    :param device_type: Netmiko device type.
    :param key_lower: True - keys lower-case, False - keys upper-case.
    :param value_lower: True - values lower-case, False - values upper-case.
    :return: Interfaces map.

    :example:
        short_to_short(key_lower=True) -> {"fa": "Fa", ...}
        short_to_short(value_lower=True) -> {"Fa": "fa", ...}
    """
    data: DStr = short_to_long(device_type=device_type)
    data = {k: k for k in data}
    intf_map_l2s: DStr = long_to_short(device_type=device_type)
    intf_map_s2l: DStr = short_to_long(device_type=device_type)
    for short_upper, long_upper in intf_map_s2l.items():
        if value := intf_map_l2s.get(long_upper):
            data[short_upper] = value
    data = _lower(data, key_lower, value_lower)
    return data


def shorts(device_type: str = "", value_lower: OBool = None) -> LStr:
    """Return short names of all interfaces.

    :param device_type: Netmiko device type.
    :param value_lower: True - values lower-case, False - values upper-case.
        Default is None, lower-case and upper-case.
    :return: Short names of all interfaces.

    :example:
        shorts() -> ["At", "BAGG", "BE", ...]
        shorts(device_type="cisco_ios") -> ["At", "Eth", "Fa", ...]
        shorts(device_type="cisco_ios", value_lower=True) -> ["at", "eth", "fa", ...]
    """
    if value_lower is None:
        upper: SStr = set(short_to_short(device_type=device_type, key_lower=False))
        lower: SStr = set(short_to_short(device_type=device_type, key_lower=True))
        return sorted(upper.union(lower))
    if value_lower:
        return sorted(short_to_short(device_type=device_type, key_lower=True))
    return sorted(short_to_short(device_type=device_type, key_lower=False))


# ============================== helper ==============================


def _overlapped(data: DStr) -> DStr:
    """Return Interfaces map long-to-short that overlapped with short key in other device_type."""
    data_ = {v: k for k, v in MAP_OTHER.items()}
    data_.update({v: k for k, v in MAP_CISCO_IOS.items()})
    data_ = {k: v for k, v in data_.items() if k not in data}
    return data_


def _lower(data: DStr, key_lower: bool, value_lower: bool) -> DStr:
    """Change key, value to lower-case.

    :param data: Interfaces map data.
    :param key_lower: True - keys lower-case, False - keys upper-case.
    :param value_lower: True - values lower-case, False - values upper-case.
    """
    if key_lower:
        data = {k.lower(): v for k, v in data.items()}
    if value_lower:
        data = {k: v.lower() for k, v in data.items()}
    return data



"""IP Protocols"""

from operator import itemgetter
from typing import Any, Tuple

from netports import helpers as h
from netports.exceptions import NetportsValueError
from netports.ports import snumbers
from netports.range import Range
from netports.static import BRIEF_ALL_I
from netports.types_ import LInt, LStr, DAny, DiAny, SStr, SInt, LTIntStr

IP_NUMBERS: DiAny = {
    0: {"number": 0, "name": "hopopt", "description": "IPv6 Hop-by-Hop Option, RFC 8200"},
    1: {"number": 1, "name": "icmp", "description": "Internet Control Message Protocol, RFC 792"},
    2: {"number": 2, "name": "igmp", "description": "Internet Group Management Protocol, RFC 1112"},
    3: {"number": 3, "name": "ggp", "description": "Gateway-to-Gateway Protocol, RFC 823"},
    4: {"number": 4, "name": "ip-in-ip", "description": "IP in IP (encapsulation), RFC 2003"},
    5: {"number": 5, "name": "st", "description": "Internet Stream Protocol, RFC 1190, RFC 1819"},
    6: {"number": 6, "name": "tcp", "description": "Transmission Control Protocol, RFC 793"},
    7: {"number": 7, "name": "cbt", "description": "Core-based trees, RFC 2189"},
    8: {"number": 8, "name": "egp", "description": "Exterior Gateway Protocol, RFC 888"},
    9: {"number": 9, "name": "igp", "description": "Interior Gateway Protocol (Cisco IGRP)"},
    10: {"number": 10, "name": "bbn-rcc-mon", "description": "BBN RCC Monitoring"},
    11: {"number": 11, "name": "nvp-ii", "description": "Network Voice Protocol, RFC 741"},
    12: {"number": 12, "name": "pup", "description": "Xerox PUP"},
    13: {"number": 13, "name": "argus", "description": "ARGUS"},
    14: {"number": 14, "name": "emcon", "description": "EMCON"},
    15: {"number": 15, "name": "xnet", "description": "Cross Net Debugger, IEN 158"},
    16: {"number": 16, "name": "chaos", "description": "Chaos"},
    17: {"number": 17, "name": "udp", "description": "User Datagram Protocol, RFC 768"},
    18: {"number": 18, "name": "mux", "description": "Multiplexing, IEN 90"},
    19: {"number": 19, "name": "dcn-meas", "description": "DCN Measurement Subsystems"},
    20: {"number": 20, "name": "hmp", "description": "Host Monitoring Protocol, RFC 869"},
    21: {"number": 21, "name": "prm", "description": "Packet Radio Measurement"},
    22: {"number": 22, "name": "xns-idp", "description": "XEROX NS IDP"},
    23: {"number": 23, "name": "trunk-1", "description": "Trunk-1"},
    24: {"number": 24, "name": "trunk-2", "description": "Trunk-2"},
    25: {"number": 25, "name": "leaf-1", "description": "Leaf-1"},
    26: {"number": 26, "name": "leaf-2", "description": "Leaf-2"},
    27: {"number": 27, "name": "rdp", "description": "Reliable Data Protocol, RFC 908"},
    28: {
        "number": 28,
        "name": "irtp",
        "description": "Internet Reliable Transaction Protocol, RFC 938",
    },
    29: {"number": 29, "name": "iso-tp4", "description": "ISO Transport Protocol Class 4, RFC 905"},
    30: {"number": 30, "name": "netblt", "description": "Bulk Data Transfer Protocol, RFC 998"},
    31: {"number": 31, "name": "mfe-nsp", "description": "MFE Network Services Protocol"},
    32: {"number": 32, "name": "merit-inp", "description": "MERIT Internodal Protocol"},
    33: {
        "number": 33,
        "name": "dccp",
        "description": "Datagram Congestion Control Protocol, RFC 4340",
    },
    34: {"number": 34, "name": "3pc", "description": "Third Party Connect Protocol"},
    35: {
        "number": 35,
        "name": "idpr",
        "description": "Inter-Domain Policy Routing Protocol, RFC 1479",
    },
    36: {"number": 36, "name": "xtp", "description": "Xpress Transport Protocol"},
    37: {"number": 37, "name": "ddp", "description": "Datagram Delivery Protocol"},
    38: {
        "number": 38,
        "name": "idpr-cmtp",
        "description": "IDPR Control Message Transport Protocol",
    },
    39: {"number": 39, "name": "tp++", "description": "TP++ Transport Protocol"},
    40: {"number": 40, "name": "il", "description": "IL Transport Protocol"},
    41: {
        "number": 41,
        "name": "ipv6",
        "description": "IPv6 Encapsulation (6to4 and 6in4), RFC 2473",
    },
    42: {"number": 42, "name": "sdrp", "description": "Source Demand Routing Protocol, RFC 1940"},
    43: {"number": 43, "name": "ipv6-route", "description": "Routing Header for IPv6, RFC 8200"},
    44: {"number": 44, "name": "ipv6-frag", "description": "Fragment Header for IPv6, RFC 8200"},
    45: {"number": 45, "name": "idrp", "description": "Inter-Domain Routing Protocol"},
    46: {"number": 46, "name": "rsvp", "description": "Resource Reservation Protocol, RFC 2205"},
    47: {
        "number": 47,
        "name": "gre",
        "description": "Generic Routing Encapsulation, RFC 2784, RFC 2890",
    },
    48: {"number": 48, "name": "dsr", "description": "Dynamic Source Routing Protocol, RFC 4728"},
    49: {"number": 49, "name": "bna", "description": "Burroughs Network Architecture"},
    50: {"number": 50, "name": "esp", "description": "Encapsulating Security Payload, RFC 4303"},
    51: {"number": 51, "name": "ah", "description": "Authentication Header, RFC 4302"},
    52: {
        "number": 52,
        "name": "i-nlsp",
        "description": "Integrated Net Layer Security Protocol, TUBA",
    },
    53: {"number": 53, "name": "swipe", "description": "SwIPe, RFC 5237"},
    54: {"number": 54, "name": "narp", "description": "NBMA Address Resolution Protocol, RFC 1735"},
    55: {"number": 55, "name": "mobile", "description": "IP Mobility (Min Encap), RFC 2004"},
    56: {
        "number": 56,
        "name": "tlsp",
        "description": "Transport Layer Security Protocol (using Kryptonet key management)",
    },
    57: {
        "number": 57,
        "name": "skip",
        "description": "Simple Key-Management for Internet Protocol, RFC 2356",
    },
    58: {"number": 58, "name": "ipv6-icmp", "description": "ICMP for IPv6, RFC 4443, RFC 4884"},
    59: {"number": 59, "name": "ipv6-nonxt", "description": "No Next Header for IPv6, RFC 8200"},
    60: {
        "number": 60,
        "name": "ipv6-opts",
        "description": "Destination Options for IPv6, RFC 8200",
    },
    61: {"number": 61, "name": "any", "description": "host internal protocol"},
    62: {"number": 62, "name": "cftp", "description": "CFTP"},
    63: {"number": 63, "name": "any", "description": "local network"},
    64: {"number": 64, "name": "sat-expak", "description": "SATNET and Backroom EXPAK"},
    65: {"number": 65, "name": "kryptolan", "description": "Kryptolan"},
    66: {"number": 66, "name": "rvd", "description": "MIT Remote Virtual Disk Protocol"},
    67: {"number": 67, "name": "ippc", "description": "Internet Pluribus Packet Core"},
    68: {"number": 68, "name": "any", "description": "distributed file system"},
    69: {"number": 69, "name": "sat-mon", "description": "SATNET Monitoring"},
    70: {"number": 70, "name": "visa", "description": "VISA Protocol"},
    71: {"number": 71, "name": "ipcu", "description": "Internet Packet Core Utility"},
    72: {"number": 72, "name": "cpnx", "description": "Computer Protocol Network Executive"},
    73: {"number": 73, "name": "cphb", "description": "Computer Protocol Heart Beat"},
    74: {"number": 74, "name": "wsn", "description": "Wang Span Network"},
    75: {"number": 75, "name": "pvp", "description": "Packet Video Protocol"},
    76: {"number": 76, "name": "br-sat-mon", "description": "Backroom SATNET Monitoring"},
    77: {"number": 77, "name": "sun-nd", "description": "SUN ND PROTOCOL-Temporary"},
    78: {"number": 78, "name": "wb-mon", "description": "WIDEBAND Monitoring"},
    79: {"number": 79, "name": "wb-expak", "description": "WIDEBAND EXPAK"},
    80: {
        "number": 80,
        "name": "iso-ip",
        "description": "International Organization for Standardization Internet Protocol",
    },
    81: {
        "number": 81,
        "name": "vmtp",
        "description": "Versatile Message Transaction Protocol, RFC 1045",
    },
    82: {
        "number": 82,
        "name": "secure-vmtp",
        "description": "Secure Versatile Message Transaction Protocol, RFC 1045",
    },
    83: {"number": 83, "name": "vines", "description": "VINES"},
    84: {"number": 84, "name": "ttp", "description": "TTP"},
    85: {"number": 85, "name": "nsfnet-igp", "description": "NSFNET-IGP"},
    86: {"number": 86, "name": "dgp", "description": "Dissimilar Gateway Protocol"},
    87: {"number": 87, "name": "tcf", "description": "TCF"},
    88: {"number": 88, "name": "eigrp", "description": "EIGRP, Informational RFC 7868"},
    89: {"number": 89, "name": "ospf", "description": "Open Shortest Path First, RFC 2328"},
    90: {"number": 90, "name": "sprite-rpc", "description": "Sprite RPC Protocol"},
    91: {"number": 91, "name": "larp", "description": "Locus Address Resolution Protocol"},
    92: {"number": 92, "name": "mtp", "description": "Multicast Transport Protocol"},
    93: {"number": 93, "name": "ax.25", "description": "AX.25"},
    94: {"number": 94, "name": "os", "description": "KA9Q NOS compatible IP over IP tunneling"},
    95: {"number": 95, "name": "micp", "description": "Mobile Internetworking Control Protocol"},
    96: {"number": 96, "name": "scc-sp", "description": "Semaphore Communications Sec. Pro"},
    97: {
        "number": 97,
        "name": "etherip",
        "description": "Ethernet-within-IP Encapsulation, RFC 3378",
    },
    98: {"number": 98, "name": "encap", "description": "Encapsulation Header, RFC 1241"},
    99: {"number": 99, "name": "any", "description": "private encryption scheme"},
    100: {"number": 100, "name": "gmtp", "description": "GMTP"},
    101: {"number": 101, "name": "ifmp", "description": "Ipsilon Flow Management Protocol"},
    102: {"number": 102, "name": "pnni", "description": "PNNI over IP"},
    103: {"number": 103, "name": "pim", "description": "Protocol Independent Multicast"},
    104: {
        "number": 104,
        "name": "aris",
        "description": "IBM ARIS (Aggregate Route IP Switching) Protocol",
    },
    105: {
        "number": 105,
        "name": "scps",
        "description": "SCPS (Space Communications Protocol Standards), SCPS-TP",
    },
    106: {"number": 106, "name": "qnx", "description": "QNX"},
    107: {"number": 107, "name": "a/n", "description": "Active Networks"},
    108: {
        "number": 108,
        "name": "ipcomp",
        "description": "IP Payload Compression Protocol, RFC 3173",
    },
    109: {"number": 109, "name": "snp", "description": "Sitara Networks Protocol"},
    110: {"number": 110, "name": "compaq-peer", "description": "Compaq Peer Protocol"},
    111: {"number": 111, "name": "ipx-in-ip", "description": "IPX in IP"},
    112: {
        "number": 112,
        "name": "vrrp",
        "description": "Virtual Router Redundancy Protocol, "
        "Common Address Redundancy Protocol (not IANA assigned), 3768",
    },
    113: {"number": 113, "name": "pgm", "description": "PGM Reliable Transport Protocol, RFC 3208"},
    114: {"number": 114, "name": "any", "description": "0-hop protocol"},
    115: {
        "number": 115,
        "name": "l2tp",
        "description": "Layer Two Tunneling Protocol Version 3, RFC 3931",
    },
    116: {"number": 116, "name": "ddx", "description": "D-II Data Exchange (DDX)"},
    117: {"number": 117, "name": "iatp", "description": "Interactive Agent Transfer Protocol"},
    118: {"number": 118, "name": "stp", "description": "Schedule Transfer Protocol"},
    119: {"number": 119, "name": "srp", "description": "SpectraLink Radio Protocol"},
    120: {"number": 120, "name": "uti", "description": "Universal Transport Interface Protocol"},
    121: {"number": 121, "name": "smp", "description": "Simple Message Protocol"},
    122: {
        "number": 122,
        "name": "sm",
        "description": "Simple Multicast Protocol, draft-perlman-simple-multicast-03",
    },
    123: {"number": 123, "name": "ptp", "description": "Performance Transparency Protocol"},
    124: {
        "number": 124,
        "name": "is-is",
        "description": "Intermediate System to Intermediate System (IS-IS) Protocol over IPv4, "
        "RFC 1142 and RFC 1195",
    },
    125: {"number": 125, "name": "fire", "description": "Flexible Intra-AS Routing Environment"},
    126: {"number": 126, "name": "crtp", "description": "Combat Radio Transport Protocol"},
    127: {"number": 127, "name": "crudp", "description": "Combat Radio User Datagram"},
    128: {
        "number": 128,
        "name": "sscopmce",
        "description": "Service-Specific Connection-Oriented Protocol in a Multilink and "
        "Connectionless Environment, ITU-T Q.2111 (1999)",
    },
    129: {"number": 129, "name": "iplt", "description": ""},
    130: {"number": 130, "name": "sps", "description": "Secure Packet Shield"},
    131: {
        "number": 131,
        "name": "pipe",
        "description": "Private IP Encapsulation within IP, "
        "Expired I-D draft-petri-mobileip-pipe-00.txt",
    },
    132: {
        "number": 132,
        "name": "sctp",
        "description": "Stream Control Transmission Protocol, RFC 4960",
    },
    133: {"number": 133, "name": "fc", "description": "Fibre Channel"},
    134: {
        "number": 134,
        "name": "rsvp-e2e-ignore",
        "description": "Reservation Protocol (RSVP) End-to-End Ignore, RFC 3175",
    },
    135: {
        "number": 135,
        "name": "mobility",
        "description": "Header, Mobility Extension Header for IPv6, RFC 6275",
    },
    136: {
        "number": 136,
        "name": "udplite",
        "description": "Lightweight User Datagram Protocol, RFC 3828",
    },
    137: {
        "number": 137,
        "name": "mpls-in-ip",
        "description": "Multiprotocol Label Switching Encapsulated in IP, RFC 4023, RFC 5332",
    },
    138: {"number": 138, "name": "manet", "description": "MANET Protocols, RFC 5498"},
    139: {"number": 139, "name": "hip", "description": "Host Identity Protocol, RFC 5201"},
    140: {
        "number": 140,
        "name": "shim6",
        "description": "Site Multihoming by IPv6 Intermediation, RFC 5533",
    },
    141: {
        "number": 141,
        "name": "wesp",
        "description": "Wrapped Encapsulating Security Payload, RFC 5840",
    },
    142: {"number": 142, "name": "rohc", "description": "Robust Header Compression, RFC 5856"},
    143: {
        "number": 143,
        "name": "ethernet",
        "description": "IPv6 Segment Routing "
        "(TEMPORARY - registered 2020-01-31, expired 2021-01-31)",
    },
}
IP_NAMES_: DAny = {d["name"]: d for i, d in IP_NUMBERS.items()}
IP_ALIASES = {
    "ipinip": {"number": 4, "name": "ipinip", "description": "IP in IP (encapsulation), RFC 2003"},
    "iptm": {"number": 84, "name": "iptm", "description": "Internet Protocol Traffic Manager"},
}
IP_NAMES = {**IP_NAMES_, **IP_ALIASES}

MIN_NUMBER = 0
MAX_NUMBER = 255
ALL_NUMBERS_L = list(range(MIN_NUMBER, MAX_NUMBER + 1))
ALL_NUMBERS_S = f"{MIN_NUMBER}-{MAX_NUMBER}"


# noinspection PyIncorrectDocstring
def iip(items: Any = "", **kwargs) -> LInt:
    """Sorting IP protocol numbers and removing duplicates.

    :param items: Range of IP protocol numbers, can be unsorted and with duplicates,
        "ip" - Return all IP protocol numbers: [0, 1, ..., 255].
    :type items: str or List[int] or List[str]

    :param verbose: True - all protocol numbers in verbose mode: [0, 1, ..., 255],
        False - all protocol numbers in brief mode: [-1] to save RAM (default).
    :type verbose: bool

    :param strict: True - Raises NetportsValueError, if the protocol is unknown (default),
        False - Skips unknown protocols.

    :param all: True - Return all IP protocol numbers: [0, 1, ..., 255].
    :type all: bool

    :return: List[int] of unique sorted IP protocol numbers.
    :rtype: List[int]

    :raises NetportsValueError: If IP protocol numbers are outside valid range 0...255.

    :example: Converts mix of ip protocol names and numbers to the numbers.
        iip("icmp,tcp,7,255") -> [1, 6, 7, 255]

    :example: Converts ip to full range of ip protocol numbers.
        iip("ip") -> [0, 1, ... 254, 255]
    """
    if h.is_all(**kwargs):
        if h.is_brief(**kwargs):
            return [BRIEF_ALL_I]
        return ALL_NUMBERS_L.copy()
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return [BRIEF_ALL_I]

    items_ = [s.lower() for s in h.split(items)]

    # 0-255
    if ALL_NUMBERS_S in items_:
        if h.is_brief(**kwargs):
            return [BRIEF_ALL_I]
        return ALL_NUMBERS_L.copy()

    # names, numbers
    pairs, invalid = ip_pairs(items=items_, **kwargs)
    if h.is_strict(**kwargs):
        if invalid:
            raise NetportsValueError(f"{invalid=}")
    if not pairs:
        return []
    numbers, _ = [list(t) for t in zip(*pairs)]

    if h.is_brief(**kwargs):
        if numbers == ALL_NUMBERS_L:
            return [BRIEF_ALL_I]
    return numbers


# noinspection PyIncorrectDocstring
def sip(items: Any = "", **kwargs) -> str:
    """Soring string IP protocol numbers and removing duplicates.

    :param items: Range of IP protocol numbers, can be unsorted and with duplicates,
        "ip" - mean all numbers in range 0...255.
    :type items: str or List[int] or List[str]

    :param verbose: True - all protocol numbers in verbose mode: [0, 1, ..., 255],
        False - all protocol numbers in brief mode: [-1], to save RAM (default).
    :type verbose: bool

    :param strict: True - Raises NetportsValueError, if the protocol is unknown (default),
        False - Skips unknown protocols.
    :type strict: bool

    :param all: True - Return all IP protocol numbers: "0-255".
    :type all: bool

    :return: str of unique sorted IP protocol numbers.
    :rtype: str

    :raises NetportsValueError: If IP protocol numbers are outside valid range 0...255.

    :example:
        sip(["icmp", "tcp", "7", 255]) -> "1,6-7,255"
    """
    if h.is_all(**kwargs):
        return ALL_NUMBERS_S
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return ALL_NUMBERS_S

    numbers = iip(items, **kwargs)
    if h.is_brief_in_items(numbers):
        return ALL_NUMBERS_S

    return snumbers(numbers)


# noinspection PyIncorrectDocstring
def ip_pairs(items: Any, **kwargs) -> Tuple[LTIntStr, LStr]:
    """Split items to IP protocol Number, Name and undefined-invalid protocols.

    :param items: Range of IP protocols, can be unsorted and with duplicates.
    :type items: str or List[int] or List[str]

    :param verbose: True - all protocol numbers in verbose mode: [0, 1, ..., 255],
        False - all protocol numbers in brief mode: [-1], to save RAM (default).
    :type verbose: bool

    :return: *List[Tuple[int, str]]* Pairs of IP protocol number and name,
        List[str] Undefined protocol names and invalid numbers.
    :rtype: Tuple[List[Tuple[int, str]], List[str]]

    :example:
        ip_pairs("tcp,1,typo,256") -> [(1, "icmp"), (6, "tcp")], ["256", "typo"]
        ip_pairs(["tcp", 1, "typo", 256]) -> [(1, "icmp"), (6, "tcp")], ["256", "typo"]
    """
    items_: LStr = [s.lower() for s in h.split(items)]
    if "ip" in items_:
        if h.is_brief(**kwargs):
            return [(BRIEF_ALL_I, "ip")], []
        return ALL_PAIRS.copy(), []
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return [(BRIEF_ALL_I, "ip")], []

    numbers, names = _split_numbers_names(items_)
    names, undefined = _split_names_undefined(names)
    numbers, invalid = _split_numbers_invalid(numbers)
    numbers.update({d["number"] for s, d in IP_NAMES.items() if s in names})
    invalid.update(undefined)

    pairs: LTIntStr = []
    for name in sorted(names):
        number = IP_NAMES[name]["number"]
        if number not in numbers:
            continue
        numbers.remove(number)
        pairs.append((number, name))
    for number in sorted(numbers):
        name = IP_NUMBERS.get(number, {}).get("name") or ""
        pairs.append((number, name))
    pairs.sort(key=itemgetter(0))

    if h.is_brief(**kwargs):
        if pairs == ALL_PAIRS:
            pairs = [(BRIEF_ALL_I, "ip")]

    return pairs, sorted(invalid)


# ============================= helpers ==============================


def all_pairs():
    """Return all pairs of IP protocol (number, name)."""
    paris = []
    for number in ALL_NUMBERS_L:
        name = IP_NUMBERS.get(number, {}).get("name") or ""
        paris.append((number, name))
    return paris


def _split_numbers_names(items: LStr) -> Tuple[SInt, SStr]:
    """Split `items` to the numbers and names.

    :param items: Combo of numbers and names.
    :return: numbers, names.
    """
    names: SStr = set()
    numbers: SInt = set()
    for item in items:
        try:
            range_o = Range(item)
            numbers.update(range_o.numbers())
        except NetportsValueError:
            names.add(item)
    return numbers, names


def _split_names_undefined(items: SStr) -> Tuple[SStr, SStr]:
    """Split `items` to the defined names and undefined names.

    :param items: Combo of defined names and undefined names.
    :return: defined names, undefined names.
    """
    names: SStr = set()
    undefined: SStr = set()
    for item in items:
        if item in IP_NAMES_:
            names.add(item)
            continue
        if item in IP_ALIASES:
            names.add(item)
            continue
        undefined.add(item)
    return names, undefined


def _split_numbers_invalid(items: SInt) -> Tuple[SInt, SStr]:
    """Split `items` to the valid protocol numbers and invalid numbers.

    :param items: Combo of valid and invalid numbers.
    :return: valid numbers, invalid numbers.
    """
    numbers: SInt = set()
    invalid: SInt = set()
    for item in items:
        if item in ALL_NUMBERS_L:
            numbers.add(item)
        else:
            invalid.add(item)
    invalid_: SStr = {str(i) for i in invalid}
    return numbers, invalid_


ALL_PAIRS = all_pairs()



"""TCP/UDP ports."""

from typing import Any

from netports import helpers as h
from netports.exceptions import NetportsValueError
from netports.ports import inumbers, parse_range
from netports.range import Range
from netports.static import BRIEF_ALL_I
from netports.types_ import LInt

MIN_PORT = 1
MAX_PORT = 65535
ALL_PORTS_L = list(range(MIN_PORT, MAX_PORT + 1))
ALL_PORTS_S = f"{MIN_PORT}-{MAX_PORT}"


# noinspection PyIncorrectDocstring
def itcp(items: Any = "", **kwargs) -> LInt:
    """Integer TCP/UDP Ports. Sorting TCP/UDP ports and removing duplicates.

    :param items: Range of TCP/UDP ports, can be unsorted and with duplicates.
    :type items: str or List[int] or List[str]

    :param verbose: True - all ports in verbose mode: [1, 2, ..., 65535],
        False - all ports in brief mode: [-1], to save RAM (default).
    :type verbose: bool

    :param all: True - Returns all TCP/UDP ports: [1, 2, ..., 65535], or [-1] for verbose=False.
    :type all: bool

    :return: List[int] of unique sorted TCP/UDP ports.
    :rtype: List[int]

    :raises NetportsValueError: If TCP/UDP ports are outside valid range 1...65535.

    :example:
        itcp("1,3-5") -> [1, 3, 4, 5]
    """
    if h.is_all(**kwargs):
        if h.is_brief(**kwargs):
            return [BRIEF_ALL_I]
        return ALL_PORTS_L.copy()
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            return [BRIEF_ALL_I]

    ports: LInt = inumbers(items)
    check_ports(ports=ports, strict=True)

    if h.is_brief(**kwargs):
        if ports == ALL_PORTS_L:
            return [BRIEF_ALL_I]
    return ports


# noinspection PyIncorrectDocstring
def stcp(items: Any = "", **kwargs) -> str:
    """String TCP/UDP ports. Sorting TCP/UDP ports and removing duplicates.

    :param items: Range of TCP/UDP ports, can be unsorted and with duplicates.
    :type items: str or List[int] or List[str]

    :param verbose: True - all ports in verbose mode: [1, 2, ..., 65535],
        False - all ports in brief mode: [-1], to save RAM (default).
    :type verbose: bool

    :param all: True - Returns all TCP/UDP ports: "1-65535".
    :type all: bool

    :return: str of unique sorted TCP/UDP ports.
    :rtype: str

    :raises NetportsValueError: If TCP/UDP ports are outside valid range 1...65535.

    :example:
        stcp([1, 3, 4, 5]) -> "1,3-5"
    """
    if h.is_all(**kwargs):
        return ALL_PORTS_S
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            items_ = ",".join(h.lstr(h.remove_brief_items(items)))
            range_o: Range = parse_range(items_)
            check_ports(ports=range_o.numbers(), strict=True)
            return ALL_PORTS_S

    range_o = parse_range(items)
    check_ports(ports=range_o.numbers(), strict=True)
    return str(range_o)


def check_port(port: int, strict: bool = False) -> bool:
    """Check TCP/UDP port in the range 1 to 65535.

    :param int port: The TCP/UDP port that needs to be checked.
    :param bool strict: True - raise NetportsValueError if the port is invalid,
        False - return False if the port is invalid. Default is `False`.

    :return: True - If the port is in the valid range of 1 to 65535, False - otherwise.
    :rtype: bool

    :raises TypeError: If the port is not integer.
    :raises NetportsValueError: If strict=True and the port is outside the valid range.
    """
    if not isinstance(port, int):
        raise TypeError(f"{port=} {int} expected.")
    if MIN_PORT <= port <= MAX_PORT:
        return True
    if strict:
        raise NetportsValueError(f"{port=}, expected in the range {MIN_PORT} to {MAX_PORT}")
    return False


def check_ports(ports: LInt, strict: bool = False) -> bool:
    """Check TCP/UDP ports in the range 1 to 65535.

    :param List[int] ports: The TCP/UDP ports that needs to be checked.

    :param bool strict: True - raise NetportsValueError if any in the ports is invalid,
        False - return False if the port is invalid. Default is `False`.

    :return: True - if all ports is in the valid range of 1 to 65535, False - otherwise.
    :rtype: bool

    :raises TypeError: If any in the ports is not integer.
    :raises NetportsValueError: If strict=True and any in the ports is outside the valid range.
    """
    for port in ports:
        if not check_port(port=port, strict=strict):
            return False
    return True



"""IPv4 address representation in CIDR notation with host data under mask bits."""

from __future__ import annotations

import re
from functools import total_ordering
from ipaddress import IPv4Interface, IPv4Network, IPv4Address
from typing import List, Dict, Iterator

from pydantic import BaseModel, Field
from vhelpers import vre

from netports.exceptions import NetportsValueError
from netports.types_ import T2Str

RE_IP = r"\d+\.\d+\.\d+\.\d+"
RE_PREFIX = r"\d+\.\d+\.\d+\.\d+/\d+"
MASK_SPLITTER = r"[\s\\/]"


@total_ordering
class IPv4(BaseModel):
    """IPv4 address representation in CIDR notation with host data under mask bits."""

    addr: str = Field(description="IPv4 address with host data under mask bits. A.B.C.D/LEN")

    def __init__(self, *args, **kwargs):
        """Initialize IPv4.

        :param addr: IP address in CIDR notation with host data under mask bits.
        :param strict: If True, IP must be valid network address (not host address).
        :raises NetportsValueError: If the CIDR address is invalid
            or cannot be converted from network with mask format.
        """
        addr, addr_len = _validate_addr(*args, **kwargs)
        kwargs_ = {k: v for k, v in kwargs.items() if k not in ["addr", "strict"]}
        super().__init__(addr=addr, **kwargs_)
        self._interface = IPv4Interface(address=addr_len)

    def __repr__(self) -> str:
        """Representation of the object."""
        class_ = self.__class__.__name__
        return f"{class_}('{self.addr}')"

    def __str__(self) -> str:
        """String representation."""
        return str(self.addr)

    def __hash__(self) -> int:
        """Hash value of the object."""
        return hash(self._interface.network)

    def __eq__(self, other) -> bool:
        """Check if two objects are equal.

        :param other: Another object to compare.
        :return: True if objects are equal, False otherwise.
        """
        if not isinstance(other, IPv4):
            return False
        return self._interface == other._interface

    def __lt__(self, other) -> bool:
        """Compare two objects.

        :param other: Another object to compare with.
        """
        if not isinstance(other, IPv4):
            return False
        return self._interface < other._interface

    def __contains__(self, other: IPv4) -> bool:
        """Check if all IPs in the other subnet are part of this network."""
        return other._interface.network.subnet_of(self._interface.network)

    # ============================= property =============================

    @property
    def ip(self) -> str:
        """IPv4 address without prefixlen, A.B.C.D."""
        return str(self._interface.ip)

    @property
    def ip_len(self) -> str:
        """IPv4 address with prefixlen, A.B.C.D/LEN."""
        return str(self._interface)

    @property
    def net(self) -> str:
        """Network without prefixlen, A.B.C.D."""
        return str(self._interface.network.network_address)

    @property
    def len(self) -> int:
        """Prefix length."""
        return int(self._interface.network.prefixlen)

    @property
    def prefix(self) -> str:
        """IPv4 prefix without host data, A.B.C.D/LEN."""
        return str(self._interface.network)

    @property
    def representation(self) -> str:
        """Representations of the IPv4 address format.

        - cidr: CIDR notation A.B.C.D/LEN
        - host: Host address A.B.C.D
        - mask: Network mask A.B.C.D 255.255.255.0
        - wildcard: Wildcard mask A.B.C.D 0.0.0.255
        """
        if re.fullmatch(RE_PREFIX, self.addr):
            return "cidr"
        if re.fullmatch(RE_IP, self.addr):
            return "host"

        _addr, mask = vre.find2(rf"^({RE_IP}){MASK_SPLITTER}({RE_IP})$", self.addr)
        if not _addr or not mask:
            return "host"  # fallback

        mask_ = str(self.netmask)
        if mask == mask_:
            if mask == "0.0.0.0":
                return "wildcard"
            return "mask"
        return "wildcard"

    # ================================ is ================================

    @property
    def is_address(self) -> bool:
        """True if the address with host data under mask bits, False if the prefix."""
        if self.len == 32:
            return True
        try:
            IPv4Network(self.addr, strict=True)
        except ValueError:
            return True
        return False

    @property
    def is_prefix(self) -> bool:
        """True if the prefix without host data under mask bits, False if the address."""
        return not self.is_address

    @property
    def is_global(self) -> bool:
        """True if the address is defined as globally reachable iana-ipv4-special-registry."""
        return bool(self._interface.ip.is_global)

    @property
    def is_link_local(self) -> bool:
        """True if the address is reserved for link-local usage. See RFC 3927."""
        return bool(self._interface.ip.is_link_local)

    @property
    def is_loopback(self) -> bool:
        """True if this is a loopback address. See RFC 3330."""
        return bool(self._interface.ip.is_loopback)

    @property
    def is_multicast(self) -> bool:
        """True if the address is reserved for multicast use. See RFC 3171."""
        return bool(self._interface.ip.is_multicast)

    @property
    def is_private(self) -> bool:
        """True if the address is defined as not globally reachable iana-ipv4-special-registry."""
        return bool(self._interface.ip.is_private)

    @property
    def is_reserved(self) -> bool:
        """True if the address is otherwise IETF reserved."""
        return bool(self._interface.ip.is_reserved)

    @property
    def is_unspecified(self) -> bool:
        """True if the address is unspecified. See RFC 5735."""
        return bool(self._interface.ip.is_unspecified)

    # ============================= objects ==============================

    @property
    def hostmask(self) -> IPv4Address:
        """The host mask, as an IPv4Address object."""
        return self._interface.hostmask

    @property
    def netmask(self) -> IPv4Address:
        """The net mask, as an IPv4Address object."""
        return self._interface.netmask

    @property
    def network(self) -> IPv4Network:
        """The network, as an IPv4Network object."""
        return self._interface.network

    # ============================== masks ===============================

    def addr_mask(self, splitter: str = " ") -> str:
        """Address with the mask in net mask notation, A.B.C.D 255.255.255.0.

        :param splitter: String to split network address and hostmask.
        :return: Address with hostmask in net mask notation.
        """
        return f"{self._interface.ip}{splitter}{self._interface.netmask}"

    def addr_wildcard(self, splitter: str = " ") -> str:
        """Address with the mask in wildcard mask notation, A.B.C.D 0.0.0.255.

        :param splitter: String to split network address and hostmask.
        :return: Address with hostmask in wildcard mask notation.
        """
        return f"{self._interface.ip}{splitter}{self._interface.hostmask}"

    def net_mask(self, splitter: str = " ") -> str:
        """Network with the mask in net mask notation, A.B.C.D 255.255.255.0.

        :param splitter: String to split network address and hostmask.
        :return: Network address with hostmask in net mask notation.
        """
        network_address = self._interface.network.network_address
        netmask = self._interface.network.netmask
        return f"{network_address}{splitter}{netmask}"

    def net_wildcard(self, splitter: str = " ") -> str:
        """Network with the mask in wildcard mask notation, A.B.C.D 0.0.0.255.

        :param splitter: String to split network address and hostmask.
        :return: Network address with hostmask in wildcard mask notation.
        """
        network_address = self._interface.network.network_address
        hostmask = self._interface.network.hostmask
        return f"{network_address}{splitter}{hostmask}"

    # ============================== other ===============================

    def hosts(self) -> Iterator[IPv4]:
        """List of all usable hosts in the network."""
        representation = self.representation
        if representation == "wildcard":
            raise NetportsValueError(f"Invalid {representation=}")
        generator_ = self._interface.network.hosts()
        for address in generator_:
            if representation == "cidr":
                addr = f"{address}/{self.len}"
            elif representation == "host":
                addr = str(address)
            elif representation == "mask":
                addr = f"{address} {self.netmask}"
            else:
                raise NetportsValueError(f"Invalid {representation=}")
            yield IPv4(addr)


LIPv4 = List[IPv4]
DIPv4 = Dict[str, IPv4]


def _validate_addr(*args, **kwargs) -> T2Str:
    """Validate IPv4 address, convert address or network with mask to CIDR A.B.C.D/LEN format.

    :param args: Arguments containing the address value.
    :param addr: IP address in CIDR notation with host data under mask bits.
    :param strict: If True, IP must be valid network address (not host address).
    :return: IPv4 address representation in CIDR notation with host data under mask bits.
    :raises NetportsValueError: If the address is invalid
        or cannot be converted from network/mask format.
    """
    # addr
    addr = ""
    if args:
        addr = str(args[0])
    if not addr:
        addr = str(kwargs.get("addr") or "")

    # addr_len
    if re.search(rf"^{RE_PREFIX}$", addr):
        addr_len = addr
    elif re.search(rf"^{RE_IP}$", addr):
        addr_len = f"{addr}/32"
    else:
        _addr, mask = vre.find2(rf"^({RE_IP}){MASK_SPLITTER}({RE_IP})$", addr)
        if not (_addr and mask):
            raise NetportsValueError("Invalid IP address format")
        network = IPv4Network(f"0.0.0.0/{mask}")
        addr_len = f"{_addr}/{network.prefixlen}"

    # strict
    strict = bool(kwargs.get("strict"))
    _ = IPv4Network(addr_len, strict=strict)

    return addr, addr_len



"""IntfGM - Interfaces Group Manager."""

from vhelpers import vre

from netports.intf import ULIntf, LIntf, Intf, LStr
from netports.range import Range
from netports.types_ import DSStr


class IntfGM:
    """IntfGM - Interfaces Group Manager."""

    def __init__(self, items):
        """Init IntfGM

        :param items: List of Interfaces.
        :type items: str or List[str] or Intf or List[Intf]
        """
        self.items = items

    def __repr__(self):
        class_ = self.__class__.__name__
        items = len(self.items)
        return f"<{class_}: {items=}>"

    # =========================== property ===========================

    @property
    def items(self) -> LIntf:
        """Interface lin."""
        return self._items

    @items.setter
    def items(self, items: ULIntf) -> None:
        if not items:
            _items = []
        elif isinstance(items, (list, set, tuple)):
            _items = list(items)
        else:
            _items = [items]

        items_: LIntf = []
        for item in _items:
            if isinstance(item, Intf):
                items_.append(item)
            else:
                item = str(item)
                items_.append(Intf(item))
        self._items = sorted(items_)

    # =========================== methods ===========================

    def ranges(self, fmt: str = "long") -> LStr:
        """Convert interfaces names to shorted range notation.

        :param fmt: Format option:
            "cisco" - Cisco compatible format
            "long"  - Long names: ["interface Ethernet1/1-3"]
            "short" - Short names: ["Eth1/1/1-3"]
        :type fmt: str

        :return: Interfaces range notation.
        :rtype: List[str]
        """
        if fmt == "long":
            return self._ranges__long()
        if fmt == "short":
            return self._ranges__short()
        expected = ["long", "short"]
        raise ValueError(f"{fmt=} {expected=}")

    # =========================== helpers ============================

    def _ranges__long(self) -> LStr:
        """Convert interfaces names to long lines.

        :example:
            intfs = ["interface Ethernet1/1", "interface Ethernet1/2", "interface Ethernet1/3"]
            intf_gm = IntfGM(intfs)
            intf_gm.ranges() -> ["interface Ethernet1/1-3"]
        """
        items = sorted(self.items)

        data: DSStr = {}
        for intf in items:
            idx = intf.last_idx()
            part = intf.part_before(idx)
            data[part] = set()

        for intf in items:
            idx = intf.last_idx()
            part_ = intf.part_before(idx)
            id_ = intf.ids[idx]
            data[part_].add(id_)

        ranges = []
        for key, ids in data.items():
            ids_ = [int(s) for s in ids]
            range_o = Range(ids_)
            for item in range_o.items:
                ranges.append(f"{key}{item}")
        return ranges

    def _ranges__short(self) -> LStr:
        """Convert interfaces names to short lines.

        :example:
            intfs = ["interface Ethernet1/1", "interface Ethernet1/2", "interface Ethernet1/3"]
            intf_gm = IntfGM(intfs)
            intf_gm.ranges() -> ["Eth1/1-3"]
        """
        long_intfs: LStr = self._ranges__long()
        short_intfs: LStr = []
        for long_intf in long_intfs:
            long_intf = long_intf.lower()
            intf_o = Intf(long_intf)
            name_short = intf_o.name_short()
            postfix = long_intf.replace(intf_o.line, "")
            short_intf = name_short + postfix
            short_intfs.append(short_intf)
        return short_intfs


# ============================ functions =============================


def generate_intfs(start: str, end: str, base: str = "") -> LIntf:
    """Generate list of Intf objects from start to end.

    :param start: First interface name in the range.
    :type start: str

    :param end: Last interface name in the range.
    :type end: str

    :param base: Prefix of the interface name that needs to be added to each interface.
    :type base: str

    :return: List of generated Intf objects.
    :rtype: List[Intf]

    :example:
        generate_intfs(start="1/1", end="1/3") -> [Intf("1/1"), Intf("1/2"), Intf("1/3")]
    """
    names: LStr = generate_names(start=start, end=end, base=base)
    intfs: LIntf = [Intf(s) for s in names]
    return intfs


def generate_names(start: str, end: str, base: str = "") -> LStr:
    """Generate list of interface names from start to end.

    :param start: First interface name in the range.
    :type start: str

    :param end: Last interface name in the range.
    :type end: str

    :param base: Prefix of the interface name that needs to be added to each interface.
    :type base: str

    :return: List of generated interface names.
    :rtype: List[str]

    :example:
        generate_names(start="Eth1/1", end="Eth1/3") -> ["Eth1/1", "Eth1/2", "Eth1/3"]
    """
    # start
    base_first, first = vre.find2(r"(.*?)(\d+)$", start)
    if not first:
        raise ValueError("Invalid parameter `start`, digit is expected.")
    # end
    base_last, last = vre.find2(r"(.*?)(\d+)$", end)
    if not last:
        raise ValueError("Invalid parameter `end`, digit is expected.")
    # base
    if base_first != base_last:
        raise ValueError("Invalid parameters `start` and `end`. The same base name is expected.")

    names: LStr = []

    for idx in range(int(first), int(last) + 1):
        name = f"{base}{base_first}{idx}"
        names.append(name)

    return names


def names_to_range(names: LStr, fmt: str = "long") -> str:
    """Join list of interface names to range.

    :param names: List of interface names.
    :type names: List[str]

    :param fmt: Format option:
        "long"  - Long names: ["interface Ethernet1/1-3"]
        "short" - Short names: ["Eth1/1/1-3"]
    :type fmt: str

    :return: Interface range.
    :rtype: str

    :example:
        names_to_range(["Ethernet1/1", "Ethernet1/2", "Ethernet1/3"], "long") -> "Ethernet1/1-3"
        names_to_range(["Ethernet1/1", "Ethernet1/2", "Ethernet1/3"], "short") -> "Eth1/1-3"
    """
    intf_gm = IntfGM(names)
    ranges = intf_gm.ranges(fmt=fmt)
    return ",".join(ranges)


def range_to_intfs(line: str, base: str = "") -> LIntf:
    """Split interface range to list of Intf objects.

    :param line: Range of interfaces that need to be split.
    :type line: str

    :param base: Prefix of the interface name that needs to be added to each interface.
    :type base: str

    :return: List of Intf objects.
    :rtype: List[Intf]

    :example:
        range_to_intfs("1/1-3,1/5") -> [Intf("1/1"), Intf("1/2"), Intf("1/3"), Intf("1/5")]
    """
    names: LStr = []

    splitter = "-"
    items: LStr = [s.strip() for s in line.split(",")]
    items = [s for s in items if s]

    for item in items:
        if item.find(splitter) == -1:
            names.append(item)
            continue

        first_last = item.split("-")
        if len(first_last) != 2:
            raise ValueError(f"Invalid range {item=}.")
        base_first, base_last = first_last

        if base_last.isdigit():
            last = base_last
            base1, first = vre.find2(r"^(.*?)(\d+)$", base_first)
        else:
            base1, first = vre.find2(r"^(.*?)(\d+)$", base_first)
            base2, last = vre.find2(r"^(.*?)(\d+)$", base_last)
            if base1 != base2:
                raise ValueError(f"Invalid range {item=}.")

        if not (first and last):
            raise ValueError(f"Invalid range {item=}.")

        for idx in range(int(first), int(last) + 1):
            name = f"{base1}{idx}"
            names.append(name)

    intfs: LIntf = sorted([Intf(f"{base}{s}") for s in set(names)])
    return intfs


def range_to_names(line: str, base: str = "") -> LStr:
    """Split interface range to list of interface names.

    :param line: Range of interfaces that need to be split.
    :type line: str

    :param base: Prefix of the interface name that needs to be added to each interface.
    :type base: str

    :return: List of interface names.
    :rtype: List[str]

    :example:
        range_to_names("1/1-3,1/5") -> ["1/1", "1/2", "1/3", "1/5"]
        range_to_names("1/1-1/3,1/5") -> ["1/1", "1/2", "1/3", "1/5"]
    """
    intfs: LIntf = range_to_intfs(line=line, base=base)
    return [o.line for o in intfs]



"""VLAN IDs."""

from typing import Any

from netports import Range
from netports import helpers as h
from netports.exceptions import NetportsValueError
from netports.ports import inumbers, parse_range
from netports.static import BRIEF_ALL_I, RANGE_SPLITTER, SPLITTER
from netports.types_ import LInt

MIN_VLAN = 1
MAX_VLAN = 4094
ALL_VLANS_L = list(range(MIN_VLAN, MAX_VLAN + 1))
ALL_VLANS_S = f"{MIN_VLAN}-{MAX_VLAN}"
SPLITTER_HPE = " "
RANGE_SPLITTER_HPE = " to "


# noinspection PyIncorrectDocstring
def ivlan(items: Any = "", **kwargs) -> LInt:
    """Sorting integer VLAN IDs and removing duplicates.

    :param items: Range of VLANs, can be unsorted and with duplicates.
    :type items: str or List[int] or List[str]

    :param verbose: True - all VLAN IDs in verbose mode: [1, 2, ..., 4094],
        False - all VLAN IDs in brief mode: [-1], to save RAM (default).
    :type verbose: bool

    :param all: True - Returns all VLAN IDs: [1, 2, ..., 4094], or [-1] for verbose=False.
    :type all: bool

    :param splitter: Separator character between items (default ",").
    :type splitter: str

    :param range_splitter: Separator between min and max numbers in range (default "-").
    :type range_splitter: str

    :param platform: Set `splitter` and `range_splitter` to platform specific values.
        platform    splitter    range_splitter  description
        ==========  ==========  ==============  ==========================
        "cisco"     ","         "-"             Cisco IOS
        "hpe"       " "         " to "          Hewlett Packard Enterprise
    :type platform: str

    :return: List[int] of unique sorted VLANs.
    :rtype: List[int]

    :raises NetportsValueError: if VLANs are outside valid range 1...4094.

    :example:
        ivlan("1,3-5") -> [1, 3, 4, 5]
    """
    if h.is_all(**kwargs):
        if h.is_brief(**kwargs):
            return [BRIEF_ALL_I]
        return ALL_VLANS_L.copy()
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            items_ = h.remove_brief_items(items)
            ports = inumbers(items_)
            check_vlans(ports)
            return [BRIEF_ALL_I]

    kwargs = _update_splitters(**kwargs)
    vlans: LInt = inumbers(items, **kwargs)
    check_vlans(vlans)

    if h.is_brief(**kwargs):
        if vlans == ALL_VLANS_L:
            return [BRIEF_ALL_I]
    return vlans


# noinspection PyIncorrectDocstring
def svlan(items: Any = "", **kwargs) -> str:
    """Sorting string VLANs and removing duplicates.

    :param items: Range of VLANs, can be unsorted and with duplicates.
    :type items str or List[int] or List[str]

    :param verbose: True - all VLAN IDs in verbose mode: [1, 2, ..., 4094],
        False - all VLAN IDs in brief mode: [-1], to save RAM (default).
    :type verbose: bool

    :param all: True - Returns all VLAN IDs: "1-4094".
    :type all: bool

    :param splitter: Separator character between items (default ",").
    :type splitter: str

    :param range_splitter: Separator between min and max numbers in range (default "-").
    :type range_splitter: str

    :param platform: Set `splitter` and `range_splitter` to platform specific values.
        platform    splitter    range_splitter  description
        ==========  ==========  ==============  ==========================
        "cisco"     ","         "-"             Cisco IOS
        "hpe"       " "         " to "          Hewlett Packard Enterprise
    :type platform: str

    :return: str of unique sorted VLANs.
    :rtype: str

    :raises NetportsValueError: if VLANs are outside valid range 1...4094.

    :example:
        svlan([1, 3, 4, 5]) -> "1,3-5"
    """
    kwargs = _update_splitters(**kwargs)
    if h.is_all(**kwargs):
        return _replace_range_splitter(ALL_VLANS_S, **kwargs)
    if h.is_brief(**kwargs):
        if h.is_brief_in_items(items):
            items_ = ",".join(h.lstr(h.remove_brief_items(items)))
            range_o: Range = parse_range(items_)
            check_vlans(range_o.numbers())
            return _replace_range_splitter(ALL_VLANS_S, **kwargs)

    range_o = parse_range(items, **kwargs)
    check_vlans(range_o.numbers())
    return str(range_o)


# ============================= helpers ==============================


def _replace_range_splitter(item: str, **kwargs) -> str:
    """Replace "-" to range_splitter specified in kwargs."""
    range_splitter = kwargs.get("range_splitter") or ""
    if not range_splitter:
        return item
    item = item.replace("-", range_splitter)
    return item


def _update_splitters(**kwargs):
    """Update `splitter` and `range_splitter` by `platform`.

    platform    splitter    range_splitter  description
    ==========  ==========  ==============  ==========================
    "cisco"     ","         "-"             Cisco IOS
    "hpe"       " "         " to "          Hewlett Packard Enterprise
    """
    platform = str(kwargs.get("platform") or "")
    if platform == "cisco":
        return {**kwargs, **{"splitter": SPLITTER, "range_splitter": RANGE_SPLITTER}}
    if platform == "hpe":
        return {**kwargs, **{"splitter": SPLITTER_HPE, "range_splitter": RANGE_SPLITTER_HPE}}
    return kwargs


# ============================= functions ==============================


def check_vlans(items: LInt) -> bool:
    """Check VLAN IDs.

    :param items: VLAN IDs.
    :return: True if all items are in the valid VLAN range 1...4094.
    :raises NetportsValueError: If on of item is outside valid range.
    """
    if invalid_vlan := [i for i in items if i < MIN_VLAN or i > MAX_VLAN]:
        raise NetportsValueError(f"{invalid_vlan=}, expected in range 1...4094")
    return True



"""Item, element of Range."""

import re
from functools import total_ordering
from typing import List

from netports.exceptions import NetportsValueError


@total_ordering
class Item:
    """Item, element of Range."""

    def __init__(self, line: str):
        """Init Item.

        :param line: Item in str format.
        :type line: str

        :example:
            item = Item("1-3")
            item.line == "1-3"
            item.min == 1
            item.max == 3
            item.range == range(1, 3)
        """
        self.line = line

    def __repr__(self) -> str:
        """Representation of the object."""
        class_ = self.__class__.__name__
        return f"{class_}({self.line!r})"

    def __str__(self) -> str:
        """String representation."""
        return self.line

    def __hash__(self) -> int:
        """Hash value of the object."""
        return hash((self.min, self.max))

    def __eq__(self, other) -> bool:
        """Check if two objects are equal.

        :param other: Another object to compare.
        :return: True if objects are equal, False otherwise.
        """
        if self.__class__ != other.__class__:
            return False
        return self.__hash__() == other.__hash__()

    def __lt__(self, other) -> bool:
        """Compare two objects.

        :param other: Another object to compare with.
        """
        if self.__class__ != other.__class__:
            return self.line < str(other)
        if self.min != other.min:
            return self.min < other.min
        return self.max < other.max

    # =========================== property ===========================

    @property
    def line(self):
        """Item in str format."""
        return self._line

    @line.setter
    def line(self, line: str) -> None:
        if not re.match(r"\d+(-\d+)?$", line):
            raise NetportsValueError(f"{line=}, expected range")
        items = line.split("-")
        self._min = int(items[0])
        self._max = self._min if len(items) == 1 else int(items[1])
        if self.min > self.max:
            raise NetportsValueError(f"{self.max=} < {self.min=}")
        self._line = line

    @property
    def min(self):
        """First int in range."""
        return self._min

    @property
    def max(self):
        """Last int in range."""
        return self._max

    @property
    def range(self):
        """Item in *range* format."""
        return range(self.min, self.max + 1)


LItem = List[Item]



"""Software Version."""

from packaging.version import Version
from vhelpers import vre, vstr

from netports.types_ import TStrInt


class SwVersion(Version):
    """Software Version."""

    def __init__(self, text: str):
        """Init SwVersion.
        :param text: Cisco version text: "12.2(55)SE12".
        :type text: str
        """
        self._text = self._init_name(name=text)
        version, nano = self._parse_version(self._text)
        super().__init__(version)
        self._nano: int = nano

    # ========================== redefined ===========================

    def __repr__(self) -> str:
        class_ = self.__class__.__name__
        params = vstr.repr_params(self._text)
        return f"{class_}({params})"

    def __str__(self) -> str:
        return self._text

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other) -> bool:
        if not isinstance(other, SwVersion):
            return False
        return self._text == other._text

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    # =========================== property ===========================

    @property
    def public(self) -> str:
        """Public version text."""
        return self._text

    @property
    def nano(self) -> int:
        """4th part of version.

        :example:
            version = SwVersion("12.2(55)SE14")
            version.nano -> 14
        """
        return self._nano

    # =========================== helpers ============================

    @staticmethod
    def _init_name(**kwargs) -> str:
        """Init name."""
        name = kwargs.get("name")
        if name is None:
            name = ""
        if not name:
            name = "0"
        return str(name).lower()

    @staticmethod
    def _parse_version(text: str) -> TStrInt:
        """Init SwVersion. Split `text` to *Version* and `nano` (4th digit)."""
        nano = 0
        items = list(vre.find4(r"(\d+)\D+(\d+)\D+(\d+)\D+(\d+)", text))
        if items[3]:
            nano = int(items[3])
        else:
            items = list(vre.find3(r"(\d+)\D+(\d+)\D+(\d+)", text))
        if not items[0]:
            items = list(vre.find2(r"(\d+)\D+(\d+)", text))
            if not items[0]:
                if version := vre.find1(r"(\d+)", text):
                    items = [version]
                else:
                    items = []
        version = ".".join(items)
        if not version:
            version = "0"
        return version, nano



"""Typing"""

from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union

# 1 level
DAny = Dict[str, Any]
DStr = Dict[str, str]
DiAny = Dict[int, Any]
IInt = Iterable[int]
IStr = Iterable[str]
LAny = List[Any]
LInt = List[int]
LStr = List[str]
OBool = Optional[bool]
OInt = Optional[int]
SInt = Set[int]
SStr = Set[str]
StrInt = Union[str, int]
T2Str = Tuple[str, str]
T3Str = Tuple[str, str, str]
T4Str = Tuple[str, str, str, str]
T5Str = Tuple[str, str, str, str, str]
T6Str = Tuple[str, str, str, str, str, str]
T7Str = Tuple[str, str, str, str, str, str, str]
TIntStr = Tuple[int, str]
TStrInt = Tuple[str, int]

# 2 level
DLStr = Dict[str, LStr]
DSStr = Dict[str, SStr]
IStrInt = Union[IStr, IInt]
LT2Str = List[T2Str]
LTIntStr = List[TIntStr]
OLStr = Optional[LStr]
StrIInt = Union[str, int, IInt]
T2SInt = Tuple[SInt, SInt]

# 3 level
OLT2Str = Optional[LT2Str]
